# import typing

# from alang.funcs import Function
# from alang.nodes import AccessMode, AddressSpace, CodeOptions, Module, Variable
# from alang.typs import Array, Struct, Vector, Tensor, int_type, float_type, tensor_type
# from alang.langs import get_language
# from alang.compiler import Compiler

if __name__ == "__main__":
    pass



import re
from typing import Optional

from alang.nodes import Node, Node, NodeLink, NodeLinks, NodeType, NodeAttr

class TypeRef(Node):
    name = NodeAttr()
    def __init__(self, name: str, node_type: NodeType):
        super().__init__(node_type)
        self.name = name

class Type(TypeRef):
    def __init__(self, name: str, node_type: NodeType):
        super().__init__(name, node_type)
        self.is_primitive = False
        self.is_algebraic = False
        self.is_scalar = False
        self.is_array = False
        self.is_vector = False
        self.is_struct = False
        self.is_tensor = False
        self.is_floatish = False
        self.is_intish = False
        self.is_boolish = False
        self.is_indexable = False
        self.is_void = False
        self.resolved_type = self
    def resolve_type(self, diags: "compiler.Diagnostics") -> "Type": # type: ignore
        return self
    @property
    def layout(self) -> "TypeLayout":
        return self.get_layout(None)
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> "TypeLayout":
        raise NotImplementedError(f"Type {self.name} does not have a layout")
    def get_type_suffix(self) -> str:
        raise NotImplementedError()
    def get_flat_index(self, indices: list[int]) -> Optional[int]:
        return None
    def ptr(self, address_space: Optional[str] = None, access_mode: Optional[str] = None) -> "Pointer":
        return Pointer(self, address_space, access_mode)
    
class TypeLayout:
    def __init__(self):
        self.byte_size = 0
        self.align = 0

class ArrayLayout(TypeLayout):
    def __init__(self):
        super().__init__()
        self.element_stride = 0
        self.num_elements = 0

def get_array_layout(element_type: Type, length: Optional[int], buffer_byte_size: Optional[int] = None) -> ArrayLayout:
    e_layout = element_type.layout
    layout = ArrayLayout()
    layout.element_stride = round_up(e_layout.align, e_layout.byte_size)
    if length is None:
        if buffer_byte_size is None:
            print("Warning: Buffer size must be provided for runtime sized arrays")
            layout.num_elements = 0
        else:
            # If B is the effective buffer binding size for the binding on the
            # draw or dispatch command, the number of elements is:
            #   N_runtime = floor(B / element stride)
            layout.num_elements = buffer_byte_size // layout.element_stride
    else:
        layout.num_elements = length
    layout.byte_size = layout.element_stride * layout.num_elements
    layout.align = e_layout.align
    return layout

class Alias(TypeRef):
    aliased_type = NodeLink()
    def __init__(self, name: str, aliased_type: Type):
        super().__init__(name, NodeType.ALIAS)
        self.aliased_type = aliased_type

class Array(Type):
    element_type = NodeLink()
    num_elements = NodeAttr()
    def __init__(self, element_type: Type, length: Optional[int]):
        element_type = try_resolve_type(element_type, self)
        super().__init__(f"{element_type.name}[{length}]", NodeType.ARRAY)
        self.element_type = element_type
        self.num_elements = length
        self.nobuffer_layout = get_array_layout(element_type, length)
        self.is_array = True
        self.is_indexable = True
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> ArrayLayout:
        if buffer_byte_size is not None:
            return get_array_layout(self.element_type, self.num_elements, buffer_byte_size)
        return self.nobuffer_layout
    @property
    def layout(self) -> ArrayLayout:
        return self.get_layout(None)
    @property
    def is_fixed_size(self) -> bool:
        return self.num_elements is not None
    @property
    def is_runtime_sized(self) -> bool:
        return self.num_elements is None
    def get_type_suffix(self) -> str:
        return "A"    
    def get_flat_index(self, indices: list[int]) -> Optional[int]:
        if len(indices) != 1:
            return None
        if not all(isinstance(i, int) for i in indices):
            return None
        return indices[0]

def get_int_name(bits: int, signed: bool) -> str:
    if bits == 8:
        if signed:
            return "sbyte"
        else:
            return "byte"
    elif bits == 16:
        if signed:
            return "short"
        else:
            return "ushort"
    elif bits == 32:
        if signed:
            return "int"
        else:
            return "uint"
    elif bits == 64:
        if signed:
            return "long"
        else:
            return "ulong"
    else:
        raise ValueError(f"Invalid integer size: {bits}")
    
def get_int_layout(bits: int) -> TypeLayout:
    layout = TypeLayout()
    layout.byte_size = ((bits // 8 + 3) // 4) * 4
    layout.align = layout.byte_size
    return layout

class Primitive(Type):
    def __init__(self, name: str, node_type: NodeType):
        super().__init__(name, node_type)
        self.is_primitive = True

class Algebraic(Primitive):
    def __init__(self, name: str, node_type: NodeType):
        super().__init__(name, node_type)
        self.is_algebraic = True

class Scalar(Algebraic):
    def __init__(self, name: str, node_type: NodeType):
        super().__init__(name, node_type)
        self.is_scalar = True

class Integer(Scalar):
    bits = NodeAttr()
    signed = NodeAttr()
    def __init__(self, bits: int, signed: bool):
        super().__init__(get_int_name(bits, signed), NodeType.INTEGER)
        self.bits = bits
        self.signed = signed
        self.nobuffer_layout = get_int_layout(bits)
        self.is_intish = True
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> TypeLayout:
        return self.nobuffer_layout
    def get_type_suffix(self) -> str:
        return self.name[0]

sbyte_type = Integer(8, True)
byte_type = Integer(8, False)
short_type = Integer(16, True)
ushort_type = Integer(16, False)
int_type = Integer(32, True)
uint_type = Integer(32, False)
long_type = Integer(64, True)
ulong_type = Integer(64, False)

def get_float_name(bits: int) -> str:
    if bits == 16:
        return "half"
    elif bits == 32:
        return "float"
    elif bits == 64:
        return "double"
    else:
        raise ValueError(f"Invalid float size: {bits}")
    
def get_float_layout(bits: int) -> TypeLayout:
    layout = TypeLayout()
    layout.byte_size = ((bits // 8 + 3) // 4) * 4
    layout.align = layout.byte_size
    return layout

class Float(Scalar):
    bits = NodeAttr()
    def __init__(self, bits: int):
        super().__init__(get_float_name(bits), NodeType.FLOAT)
        self.bits = bits
        self.cached_layout = get_float_layout(bits)
        self.is_floatish = True
    def get_type_suffix(self) -> str:
        return self.name[0]
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> TypeLayout:
        return self.cached_layout

half_type = Float(16)
float_type = Float(32)
double_type = Float(64)

class FunctionType(Type):
    return_type = NodeLink()
    parameter_types = NodeLinks()
    def __init__(self, return_type, param_types):
        super().__init__(None, NodeType.FUNCTION_TYPE)
        self.return_type = return_type
        for pt in param_types:
            self.link(pt, "parameter_types")

class ModuleType(Type):
    def __init__(self, name: str):
        super().__init__(name, NodeType.MODULE_TYPE)

class Pointer(Type):
    element_type = NodeLink()
    address_space = NodeAttr()
    access_mode = NodeAttr()
    def __init__(self, element_type: Type, address_space: Optional[str] = None, access_mode: Optional[str] = None):
        element_type = try_resolve_type(element_type, self)
        super().__init__(f"{element_type.name}*", NodeType.POINTER)
        self.element_type = element_type
        self.address_space = address_space
        self.access_mode = access_mode
    def get_type_suffix(self) -> str:
        return "P"
    def resolve_type(self, diags: "compiler.Diagnostics") -> Optional[Type]: # type: ignore
        et = self.element_type
        if et is None:
            return None
        ret = et.resolved_type
        if ret is None:
            return None
        if ret.id == et.id:
            return self
        return ret.ptr(self.address_space, self.access_mode)

class Field(Node):
    name = NodeAttr()
    field_type = NodeLink()
    def __init__(self, name: str, field_type: Type):
        super().__init__(NodeType.FIELD)
        self.name = name
        self.field_type = field_type
    def resolve_type(self, diags: "compiler.Diagnostics") -> Optional[Type]: # type: ignore
        ft = self.field_type
        if ft is not None:
            return ft.resolved_type
        return None

class FieldLayout(TypeLayout):
    def __init__(self):
        self.offset = 0
        self.num_elements = 0
    @property
    def triple(self) -> tuple[int, int, int]:
        """Returns a tuple of (offset, align, size)"""
        return (self.offset, self.align, self.byte_size)

class StructLayout(TypeLayout):
    def __init__(self):
        super().__init__()
        self.fields = []

def get_struct_layout(fields: list[Field], buffer_byte_size: Optional[int] = None) -> StructLayout:
    struct_layout = StructLayout()
    last_offset = 0
    last_size = 0
    max_field_align = 0
    if len(fields) == 0:
        return struct_layout
    for field_index, field in enumerate(fields):
        type_layout = field.field_type.layout
        offset = round_up(type_layout.align, last_offset + last_size)
        field_layout = FieldLayout()
        field_layout.offset = offset
        field_layout.num_elements = 1
        if buffer_byte_size is not None:
            remaining_bytes = buffer_byte_size - offset
            if remaining_bytes < 0:
                print(f"Warning: Buffer size too small for field {field.name}")
            else:
                type_layout = field.field_type.get_layout(remaining_bytes)
                if isinstance(type_layout, ArrayLayout):
                    field_layout.num_elements = type_layout.num_elements
        field_layout.byte_size = type_layout.byte_size
        field_layout.align = type_layout.align
        struct_layout.fields.append(field_layout)
        max_field_align = max(max_field_align, type_layout.align)
        last_offset = offset
        last_size = type_layout.byte_size
    struct_layout.align = max_field_align
    just_past_last_field = struct_layout.fields[-1].offset + struct_layout.fields[-1].byte_size
    struct_layout.byte_size = round_up(struct_layout.align, just_past_last_field)
    return struct_layout

def round_up(k: int, n: int) -> int:
    # roundUp(k, n) = ⌈n ÷ k⌉ × k
    return ((n + k - 1) // k) * k

class Struct(Type):
    fields = NodeLinks()
    def __init__(self, name, *fields: Optional[list[Field]]):
        super().__init__(name, NodeType.STRUCT)
        for field in fields:
            if isinstance(field, Field):
                self.link(field, "fields")
            else:
                self.field(*field)
        self.nobuffer_layout = None
        self.is_struct = True

    def get_layout(self, buffer_byte_size: Optional[int] = None) -> StructLayout:
        if buffer_byte_size is None:
            if self.nobuffer_layout is None:
                self.nobuffer_layout = get_struct_layout(self.fields)
            return self.nobuffer_layout
        else:
            return get_struct_layout(self.fields, buffer_byte_size)

    @property
    def layout(self) -> StructLayout:
        return self.get_layout(None)

    def field(self, name: str, type: Type) -> "Struct":
        f = Field(name, try_resolve_type(type, self))
        self.link(f, "fields")
        self.nobuffer_layout = None
        return self

def get_tensor_shape_name(shape: tuple):
    sn = "x".join([str(s) for s in shape])
    return sn

def get_tensor_name(shape: tuple, element_type: Type):
    sn = get_tensor_shape_name(shape)
    return f"{element_type.name}{sn}"

def get_tensor_mm_shape(a_shape: tuple, b_shape: tuple) -> tuple:
    if len(a_shape) != 2 or len(b_shape) != 2:
        raise ValueError("Can only multiply 2D tensors")
    if a_shape[1] != b_shape[0]:
        raise ValueError("Cannot multiply tensors with incompatible shapes")
    return (a_shape[0], b_shape[1])

class Tensor(Algebraic):
    element_type = NodeLink()
    shape = NodeAttr()
    def __init__(self, shape: tuple, element_type: Type):
        element_type = try_resolve_type(element_type, None)
        if element_type is None:
            raise ValueError(f"Cannot create tensor with unresolved element type: {element_type}")
        elif not element_type.is_scalar:
            raise ValueError(f"Cannot create tensor with non-scalar element type: {element_type}")
        super().__init__(get_tensor_name(shape, element_type), NodeType.TENSOR)
        self.element_type = element_type
        self.shape = shape
        num_elements = 1
        for s in shape:
            num_elements *= s
        self.num_elements = num_elements
        self.nobuffer_layout = get_array_layout(element_type, num_elements)
        self.is_tensor = True
        self.is_floatish = element_type.is_floatish
        self.is_intish = element_type.is_intish
        self.is_boolish = element_type.is_boolish
        self.is_indexable = True
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> ArrayLayout:
        return self.nobuffer_layout
    def __matmul__(self, other: "Tensor") -> "Tensor":
        if not other.is_tensor:
            raise ValueError("Cannot multiply tensor by non-tensor")
        out_shape = get_tensor_mm_shape(self.shape, other.shape)
        return Tensor(out_shape, self.element_type)
    def get_flat_index(self, indices: list["Expression"]) -> Optional["Expression"]: # type: ignore
        if len(indices) < 1:
            return None
        if len(indices) != len(self.shape):
            return None
        from alang.exprs import Constant
        flat_index = Constant(0)
        for i, s in enumerate(self.shape):
            flat_index = flat_index * s + indices[i]
        return flat_index
    def get_support_definitions(self, defs: "compiler.SupportDefinitions"): # type: ignore
        if defs.needs(self.name):
            defs.add(self.name, [Alias(self.name, Array(self.element_type, self.num_elements))])

class TypeName(TypeRef):
    def __init__(self, name: str):
        super().__init__(name, NodeType.TYPE_NAME)

def get_vector_name(element_type: Type, size: int) -> str:
    return f"vec{size}{element_type.get_type_suffix()}"

def get_vector_layout(element_type: Type, size: int) -> TypeLayout:
    e_layout: TypeLayout = element_type.layout
    layout = TypeLayout()
    if size == 2:
        layout.byte_size = e_layout.byte_size * 2
        layout.align = layout.byte_size
    elif size == 3:
        layout.byte_size = e_layout.byte_size * 3
        layout.align = e_layout.byte_size * 4
    elif size == 4:
        layout.byte_size = e_layout.byte_size * 4
        layout.align = layout.byte_size
    else:
        raise ValueError(f"Invalid vector size: {size}")
    return layout

class Vector(Algebraic):
    element_type = NodeLink()
    size = NodeAttr()
    def __init__(self, element_type: Type, size: int):
        super().__init__(get_vector_name(element_type, size), NodeType.VECTOR)
        if size < 2 or size > 4:
            raise ValueError(f"Invalid vector size: {size}. Size must be 2, 3, or 4")
        self.element_type = element_type
        self.size = size
        self.nobuffer_layout = get_vector_layout(element_type, size)
        self.is_vector = True
        self.is_floatish = element_type.is_floatish
        self.is_intish = element_type.is_intish
        self.is_boolish = element_type.is_boolish
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> TypeLayout:
        return self.nobuffer_layout

vec2h_type = Vector(half_type, 2)
vec3h_type = Vector(half_type, 3)
vec4h_type = Vector(half_type, 4)

vec2f_type = Vector(float_type, 2)
vec3f_type = Vector(float_type, 3)
vec4f_type = Vector(float_type, 4)

vec2i_type = Vector(int_type, 2)
vec3i_type = Vector(int_type, 3)
vec4i_type = Vector(int_type, 4)

class Void(Type):
    """The void type used only for function return types"""
    def __init__(self):
        super().__init__("void", NodeType.VOID)
        self.is_void = True
    def get_layout(self, buffer_byte_size: Optional[int] = None) -> TypeLayout:
        return TypeLayout()

void_type = Void()

scalar_types = {
    sbyte_type.name: sbyte_type,
    byte_type.name: byte_type,
    short_type.name: short_type,
    ushort_type.name: ushort_type,
    int_type.name: int_type,
    uint_type.name: uint_type,
    long_type.name: long_type,
    ulong_type.name: ulong_type,

    half_type.name: half_type,
    float_type.name: float_type,
    double_type.name: double_type,
}

builtin_types = {
    void_type.name: void_type,

    vec2h_type.name: vec2h_type,
    vec3h_type.name: vec3h_type,
    vec4h_type.name: vec4h_type,
    vec2f_type.name: vec2f_type,
    vec3f_type.name: vec3f_type,
    vec4f_type.name: vec4f_type,
    vec2i_type.name: vec2i_type,
    vec3i_type.name: vec3i_type,
    vec4i_type.name: vec4i_type,    
}
builtin_types.update(scalar_types)

def try_resolve_builtin_type(name: str) -> Optional[Type]:
    if name in builtin_types:
        return builtin_types[name]
    return None

def resolve_builtin_type(name: str) -> Type:
    bt = try_resolve_builtin_type(name)
    if bt is None:
        raise ValueError(f"Unknown builtin type: {name}")
    return bt

tensor_type_re = re.compile(r"^([a-z][a-z]+)(((\d+)x)+(\d+))$", 0)

def try_resolve_tensor_type(name: str) -> Optional[Type]:
    m = tensor_type_re.match(name)
    if m is None:
        return None
    element_type_name = m.group(1)
    element_type = try_resolve_builtin_type(element_type_name)
    if element_type is None:
        return None
    shape_str = m.group(2)
    shape = tuple([int(s) for s in shape_str.split("x")])
    return Tensor(shape, element_type)

def try_resolve_type(type, context: Optional[Node]) -> Type:
    if type is None:
        return None
    elif isinstance(type, Type):
        return type
    elif isinstance(type, str):
        bt = try_resolve_builtin_type(type)
        if bt is not None:
            return bt
        tt = try_resolve_tensor_type(type)
        if tt is not None:
            return tt
        # TODO: Implement lookup in context
        return TypeName(str(type))
    else:
        raise ValueError(f"Invalid type: {type}")

def array(element_type: str, length: Optional[int] = None) -> Array:
    return Array(element_type, length)

def tensor_type(shape: tuple, element_type: str) -> Tensor:
    return Tensor(shape, try_resolve_type(element_type, None))

def struct(name: str, *fields: list) -> Struct:
    return Struct(name, *fields)



from alang.nodes import define, global_module, Module, AccessMode, AddressSpace, Variable
from alang.typs import float_type

def test_define():
    f_count = len(global_module.functions)
    n = define("f", "x")
    assert n.name == "f"
    assert len(n.parameters) == 1
    assert n.parameters[0].name == "x"
    assert len(global_module.functions) == f_count + 1

def test_has_global_module():
    assert global_module is not None
    assert global_module.name == "global"

def test_new_module():
    m = Module("test")
    m.define("f", "x")
    m.define("g", "y")
    assert m.name == "test"
    assert len(m.functions) == 2
    assert m.functions[0].name == "f"
    assert m.functions[1].name == "g"

def test_var():
    f = define("f", "x").set("y", "2*x").ret("y")
    assert len(f.variables) == 1

def test_module_var_default_address_space():
    m = Module("test")
    m.var("x", "int")
    v = m.variables[0]
    assert v.address_space == AddressSpace.PRIVATE
    assert v.access_mode == "read_write"
    assert m.wgsl_code.strip() == """
var<private> x: i32;
""".strip()

def test_module_var_default_storage_access_mode():
    m = Module("test")
    m.var("x", "int", address_space=AddressSpace.STORAGE)
    v = m.variables[0]
    assert v.address_space == AddressSpace.STORAGE
    assert v.access_mode == "read"
    assert m.wgsl_code.strip() == """
var<storage, read> x: i32;
""".strip()

def test_module_var_storage_read_write_access_mode():
    m = Module("test")
    m.var("x", "int", address_space=AddressSpace.STORAGE, access_mode=AccessMode.READ_WRITE)
    v = m.variables[0]
    assert v.address_space == AddressSpace.STORAGE
    assert v.access_mode == "read_write"
    assert m.wgsl_code.strip() == """
var<storage, read_write> x: i32;
""".strip()

def test_function_var_default_address_space():
    m = Module("test")
    m.define("f").var("x", "int")
    f: Function = m.functions[0]
    v: Variable = f.variables[0]
    assert v.address_space is not None
    assert v.access_mode == "read_write"
    assert v.address_space == "function"
    assert f.wgsl_code.strip() == """
fn f() {
    var x: i32;
}
""".strip()

def test_pointer_param():
    f = define("f").param("x", float_type.ptr(address_space=AddressSpace.STORAGE))
    assert f.parameters[0].parameter_type.name == "float*"
    assert f.wgsl_code.strip() == """
fn f(x: ptr<storage, f32>) {
}
""".strip()
    assert f.c_code.strip() == """
void f(float* x) {
}
""".strip()



from alang.typs import array

def test_is_fixed():
    a = array("int", 10)
    assert a.is_fixed_size
    assert not a.is_runtime_sized
    assert a.num_elements == 10

def test_is_runtime():
    a = array("int")
    assert not a.is_fixed_size
    assert a.is_runtime_sized
    assert a.num_elements is None

# https://www.w3.org/TR/WGSL/#array-layout-examples

def test_small_stride():
    a = array("float", 8)
    l = a.layout
    assert l.align == 4
    assert l.element_stride == 4
    assert l.byte_size == 32

def test_bigger_stride():
    a = array("vec3f", 8)
    l = a.layout
    assert l.align == 16
    assert l.element_stride == 16
    assert l.byte_size == 128

def test_runtime_sized_nobuffer():
    a = array("float")
    l = a.layout
    assert l.align == 4
    assert l.element_stride == 4
    assert l.byte_size == 0

def test_runtime_sized_floats():
    a = array("float")
    l = a.get_layout(8)
    assert l.align == 4
    assert l.element_stride == 4
    assert l.num_elements == 2
    assert l.byte_size == 8

def test_runtime_sized_float3s_not_enough_buffer():
    a = array("vec3f")
    l = a.get_layout(8)
    assert l.align == 16
    assert l.element_stride == 16
    assert l.num_elements == 0
    assert l.byte_size == 0

def test_runtime_sized_float3s():
    a = array("vec3f")
    l = a.get_layout(48)
    assert l.align == 16
    assert l.element_stride == 16
    assert l.num_elements == 3
    assert l.byte_size == 48

def test_runtime_truncation():
    # https://www.w3.org/TR/WGSL/#buffer-binding-determines-runtime-sized-array-element-count
    a = array("float")
    assert a.get_layout(1024).num_elements == 256
    assert a.get_layout(1025).num_elements == 256
    assert a.get_layout(1026).num_elements == 256
    assert a.get_layout(1027).num_elements == 256
    assert a.get_layout(1028).num_elements == 257



from alang.nodes import define
from alang.compiler import Compiler, CodeOptions

def test_infer_void_return_type():
    f = define("f")
    assert f.return_type is None
    c = Compiler(f, CodeOptions())
    c.compile()
    assert f.return_type.name == "void"

def test_infer_const_return_type():
    f = define("f").ret(42)
    assert f.return_type is None
    c = Compiler(f, CodeOptions())
    c.compile()
    assert f.return_type.name == "int"



from typing import Optional

from alang.nodes import Block, Expression, NodeAttr, NodeLink, NodeType, Statement
import alang.typs as typs

class ExprStmt(Statement):
    expression = NodeLink()
    def __init__(self, expression: Expression):
        super().__init__(NodeType.EXPR_STMT)
        self.expression = expression
    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        return self.expression.resolved_type
    
class Loop(Block):
    var = NodeAttr()
    count = NodeLink()
    def __init__(self, var: str, count: Optional[Expression], *statements: list[Statement]):
        super().__init__(NodeType.LOOP, can_define_types=False, can_define_functions=False, can_define_variables=False, can_define_statements=True)
        self.var = str(var)
        if isinstance(count, int):
            from alang.exprs import Constant
            count = Constant(count)
        self.count = count
        self.statements = statements
    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        return typs.void_type

class Return(Statement):
    value = NodeLink()
    def __init__(self, value: Optional[Expression]):
        super().__init__(NodeType.RETURN)
        self.value = value
    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        if self.value is None:
            return typs.void_type
        return self.value.resolved_type

class Set(Statement):
    target = NodeLink()
    value = NodeLink()
    def __init__(self, target: Expression, value: Expression):
        super().__init__(NodeType.SET)
        self.target = target
        self.value = value
    def resolve_type(self, diags: "compiler.Diagnostics") -> typs.Type: # type: ignore
        if self.target.resolved_type is None:
            if self.value.resolved_type is None:
                return None
            else:
                return self.value.resolved_type
        else:
            return self.target.resolved_type



from alang.nodes import define
from alang.typs import int_type
from test_html import write_standalone_html

def test_call():
    f = define("f").param("x", int_type).call("do_something", "x")
    code = f.wgsl_code
    assert code.strip() == """
fn f(x: i32) {
    do_something(x);
}
// ERROR! Name do_something not found (name (name='do_something'))""".strip()

def test_loop():
    f = define("f").loop("i", 10)
    code = f.wgsl_code
    assert code.strip() == """
fn f() {
    for (var i: i32 = 0; i < 10; i++) {
    }
}
""".strip()
    write_standalone_html("test_loop", f)



from typing import Optional

from alang.nodes import BreadthFirstVisitor, CodeOptions, DepthFirstVisitor, Node, NodeType
from alang.typs import void_type
from alang.funcs import Function

class DiagnosticKind:
    ERROR = "error"
    WARNING = "warning"

class DiagnosticMessage:
    def __init__(self, kind: str, message: str, node: Optional[Node]):
        self.message = message
        self.node = node
        self.kind = kind
    def print(self):
        if self.node is not None:
            print(f"{self.kind.upper()}: {self.message} {self.node}")
        else:
            print(f"{self.kind.upper()}: {self.message}")

class Diagnostics:
    def __init__(self):
        self.messages: list[DiagnosticMessage] = []
        self.num_errors = 0
    def reset(self):
        self.messages = []
        self.num_errors = 0
    def message(self, kind: str, message: str, node: Optional[Node] = None):
        m = DiagnosticMessage(kind, message, node)
        self.messages.append(m)
        if kind == DiagnosticKind.ERROR:
            self.num_errors += 1
        # m.print()
    def error(self, message: str, node: Optional[Node] = None):
        self.message(DiagnosticKind.ERROR, message, node)
    def warning(self, message: str, node: Optional[Node] = None):
        self.message(DiagnosticKind.WARNING, message, node)

class TypeResolutionPass(DepthFirstVisitor):
    def __init__(self, diags: Diagnostics):
        self.diags = diags
        self.num_changes = 0
        self.num_need_info = 0
        self.num_errors = 0
    def run(self, node: Node):
        self.visit(node, None, None, None)
    def visit_node(self, node: Node, parent: Node, rel: str, acc):
        if node.resolved_type is not None:
            return
        t = node.resolve_type(self.diags)
        if t is None:
            self.num_need_info += 1
        else:
            node.resolved_type = t
            self.num_changes += 1
        return super().visit_node(node, parent, rel, acc)
    
class NameResolutionPass(BreadthFirstVisitor):
    def __init__(self, diags: Diagnostics):
        self.diags = diags
        self.num_changes = 0
        self.num_need_info = 0
        self.num_errors = 0
    def run(self, node: Node):
        self.visit(node, None, None, dict())
    def visit_name(self, node: Node, parent: Node, rel: str, env: dict): # type: ignore
        if node.resolved_node is not None:
            return env
        elif node.name in env:
            node.resolved_node = env[node.name]
            self.num_changes += 1
            return env
        else:
            self.diags.error(f"Name {node.name} not found", node)
            self.num_errors += 1
            return env
    def visit_function(self, node: Function, parent: Node, rel: str, env: dict):
        new_env = dict(env)
        for p in node.parameters:
            new_env[p.name] = p
        return new_env

class InferFunctionReturnTypePass(DepthFirstVisitor):
    def __init__(self, diags: Diagnostics):
        self.diags = diags
        self.num_changes = 0
        self.num_need_info = 0
        self.num_errors = 0
    def run(self, node: Node):
        self.visit(node, None, None, None)
    def set_return_type(self, node: Function, return_type: str):
        node.return_type = return_type
        self.num_changes += 1
    def visit_function(self, node: Function, parent: Node, rel: str, cvals: list):
        if node.return_type is not None:
            return
        return_values: list[Node] = [x.value for x in node.find_reachable_with_type(NodeType.RETURN)]
        if len(return_values) == 0 or all(x is None for x in return_values):
            # No return statements, or all return statements are empty, so return void
            self.set_return_type(node, void_type)
            return
        # See if any of them have resolved types
        maybe_resolved_types = [x.resolved_type for x in return_values if x is not None]
        resolved_types = [x for x in maybe_resolved_types if x is not None]
        # Distinct them by name
        distinct_types = {}
        for t in resolved_types:
            distinct_types[t.name] = t
        if len(distinct_types) == 0:
            self.num_need_info += 1
            return
        if len(distinct_types) > 1:
            self.num_errors += 1
            self.diags.error(f"Multiple return types found: {distinct_types}", node)
        return_type = list(distinct_types.values())[0]
        self.set_return_type(node, return_type)

class SupportDefinitions:
    def __init__(self):
        self.definitions: dict[str, list[Node]] = dict()
        self.num_changes = 0
    def add(self, group: str, defs: list[Node]):
        if group in self.definitions:
            return
        self.definitions[group] = defs
        self.num_changes += 1
    def get(self, name: str):
        return self.definitions.get(name)
    def needs(self, name: str):
        return name not in self.definitions

class CollectSupportDefinitions(DepthFirstVisitor):
    def __init__(self):
        self.defs = SupportDefinitions()
    @property
    def num_changes(self):
        return self.defs.num_changes
    @property
    def support_definitions(self):
        flattened = []
        for k, v in self.defs.definitions.items():
            flattened.extend(v)
        return flattened
    def run(self, node: Node):
        self.visit(node, None, None, None)
    def visit_node(self, node: Node, parent: Node, rel: str, acc):
        node.get_support_definitions(self.defs)
        super().visit_node(node, parent, rel, acc)

class Compiler:
    def __init__(self, ast: Node, options: CodeOptions):
        self.ast = ast
        self.options = options
        self.diags = Diagnostics()
        self.support_definitions = []
        self.entry_points = []

    def resolve_names(self):
        res_pass = NameResolutionPass(self.diags)
        res_pass.run(self.ast)
        return res_pass
    
    def resolve_types(self):
        res_pass = TypeResolutionPass(self.diags)
        res_pass.run(self.ast)
        return res_pass
    
    def infer_types(self):
        func_return_type_pass = InferFunctionReturnTypePass(self.diags)
        func_return_type_pass.run(self.ast)
        return func_return_type_pass
    
    def get_support_definitions(self):
        c = CollectSupportDefinitions()
        c.run(self.ast)
        return c.support_definitions
    
    def find_entry_points(self):
        from alang import exprs, stmts
        self.entry_points.clear()
        funcs: list[Function] = self.ast.find_reachable_with_type(NodeType.FUNCTION)
        for node in funcs:
            if node.stage is not None:
                self.entry_points.append((node, node.stage, None))
        if self.options.auto_entry_points and len(self.entry_points) == 0 and len(funcs) > 0:
            # Synthesize an entry point for the last function
            f = funcs[-1]
            ft = f.resolved_type
            if ft is None:
                self.diags.warning(f"Function {f.name}'s type could not be determined so it will not be used as an entry point", f)
            new_name = f.name + "_auto_compute"
            new_f = Function(new_name, void_type)
            new_f.stage = "compute"
            for pi, p in enumerate(f.parameters):
                new_f.param(p.name, ft.parameter_types[pi])
            # add outputs to ps
            call = exprs.Funcall(f, [p.name for p in f.parameters])
            # if not ft.return_type.is_void:
            #     new_f.param("_auto_out", ft.return_type).set("_auto_out", call)
            # else:
            new_f.stmt(stmts.ExprStmt(call))
            new_f.workgroup_size = 1
            self.entry_points.append((f, "compute", new_f))
    
    def compile(self):
        self.support_definitions.clear()
        should_iter = True
        max_iterations = 100
        iteration = 0
        while should_iter and iteration < max_iterations:
            self.diags.reset()
            resolve_name_info = self.resolve_names()
            resolve_types_info = self.resolve_types()
            infer_types_info = self.infer_types()
            should_iter = resolve_name_info.num_changes > 0 or resolve_types_info.num_changes > 0 or infer_types_info.num_changes > 0
            iteration += 1
        if iteration == max_iterations:
            self.diags.error("Max iterations reached")
        self.support_definitions = self.get_support_definitions()
        self.find_entry_points()



from typing import Optional, TextIO, Union

from alang.langs.language import Language, register_language
from alang.langs.writer import CodeWriter
import alang.exprs as exprs
import alang.funcs as funcs
import alang.nodes as nodes
import alang.stmts as stmts
import alang.typs as typs

class CWriter(CodeWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)

    def write_alias(self, a: "Alias"): # type: ignore
        self.write(f"typedef ")
        self.write_type_ref(a.aliased_type)
        self.write(f" {a.name};\n")

    def write_binop(self, b: exprs.Binop):
        support_name = b.get_support_lib_function_name()
        if support_name is not None:
            self.write(support_name)
            self.write("(")
            self.write_expr(b.left)
            self.write(", ")
            self.write_expr(b.right)
            self.write(")")
        else:
            self.write("(")
            self.write_expr(b.left)
            self.write(" ")
            self.write(b.operator.op)
            self.write(" ")
            self.write_expr(b.right)
            self.write(")")

    def write_constant(self, c: "Constant"): # type: ignore
        self.write(repr(c.value))

    def write_expr_stmt(self, e: stmts.ExprStmt):
        self.write_expr(e.expression)
        self.write(";\n")

    def write_funcall(self, c: exprs.Funcall):
        f = c.func
        if f is None:
            self.write_error_value("Missing function")
            return
        if f.node_type == nodes.NodeType.FUNCTION:
            self.write(f.name)
        else:
            self.write_expr(f)
        self.write("(")
        for i, arg in enumerate(c.args):
            if i > 0:
                self.write(", ")
            self.write_expr(arg)
        self.write(")")

    def write_function(self, f: funcs.Function):
        self.write_type_ref(f.return_type)
        self.write(f" {f.name}(")
        ps = f.parameters
        for i, param in enumerate(ps):
            self.write_type_ref(param.parameter_type)
            self.write(f" {param.name}")
            if i < len(ps) - 1:
                self.write(", ")
        self.writeln(") {")
        self.indent()
        for s in f.statements:
            self.write_node(s)
        self.dedent()
        self.writeln("}")

    def write_index(self, i: "Index"): # type: ignore
        self.write_expr(i.base)
        self.write("[")
        for ri, r in enumerate(i.ranges):
            if ri > 0:
                self.write(", ")
            if r is None:
                self.write(":")
            else:
                self.write_expr(r)
        self.write("]")

    def write_line_comment(self, comment: str):
        self.write("// ")
        self.write(comment)
        self.write("\n")

    def write_loop(self, f: stmts.Loop):
        self.write("for (")
        self.write_type_ref(typs.int_type)
        self.write(f" {f.var} = 0; {f.var} < ")
        self.write_expr(f.count)
        self.writeln(f"; {f.var}++) {{")
        self.indent()
        for s in f.statements:
            self.write_node(s)
        self.dedent()
        self.writeln("}")

    def write_return(self, r: stmts.Return):
        self.write("return")
        if r.value is not None:
            self.write(" ")
            self.write_expr(r.value)
        self.write(";\n")

    def write_name(self, n: exprs.Name):
        self.write(n.name)

    def write_set(self, s: stmts.Set):
        self.write_expr(s.target)
        self.write(" = ")
        self.write_expr(s.value)
        self.write(";\n")

    def write_type_ref(self, t: typs.Type):
        self.write(self.get_type_name(t))

    def get_type_name(self, t: typs.Type) -> str:
        if t is None:
            return "void"
        elif isinstance(t, typs.Integer):
            tn = t.name
            if tn == "sbyte":
                return "int8_t"
            elif tn == "byte":
                return "uint8_t"
            elif tn == "int":
                return "int32_t"
            elif tn == "uint":
                return "uint32_t"
            elif tn == "long":
                return "int64_t"
            elif tn == "ulong":
                return "uint64_t"
            else:
                raise ValueError(f"Invalid integer type: {tn}")
        elif isinstance(t, typs.Float):
            tn = t.name
            if tn == "half":
                return "half"
            elif tn == "float":
                return "float"
            elif tn == "double":
                return "double"
            else:
                raise ValueError(f"Invalid float type: {tn}")
        elif isinstance(t, typs.Vector):
            n = t.size
            if n > 4 or n < 2:
                raise ValueError(f"Invalid vector size: {n}")
            element_type = self.get_type_name(t.element_type)
            return f"{element_type}{n}"
        else:
            return t.name


class CLanguage(Language):
    def __init__(self):
        super().__init__("c")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> CWriter: # type: ignore
        return CWriter(out, options)

c_lang = CLanguage()
register_language(c_lang)




"""GLSL https://www.khronos.org/opengl/wiki/OpenGL_Shading_Language"""

from typing import Optional, TextIO, Union

from alang.langs.language import Language, register_language
from alang.langs.writer import CodeWriter

import alang.stmts as stmts
import alang.typs as typs
import alang.funcs as funcs

def encode(str):
    return str.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

class HTMLWriter(CodeWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)
        self.out_langs = ["wgsl", "js", "c", "metal", "glsl", "a"]

    def write_expr_stmt(self, e: stmts.ExprStmt):
        self.write_expr(e.expression)

    def write_index(self, i: "Index"): # type: ignore
        self.write_expr(i.base)
        self.write("[")
        for ri, r in enumerate(i.ranges):
            if ri > 0:
                self.write(", ")
            if r is None:
                self.write(":")
            else:
                self.write_expr(r)
        self.write("]")

    def write_line_comment(self, comment: str):
        self.write("<p>")
        self.write(encode(comment))
        self.write("</p>\n")

    def write_module(self, s: "mods.Module"): # type: ignore
        self.write(f"<html>\n")
        self.write(f"<head>\n")
        self.write(f"<title>{encode(s.name)}</title>\n")
        self.write(f"<style>\n")
        # Enable automatic dark mode
        self.write(f"html {{ color-scheme: light dark; font-family:Helvetica }}\n")
        self.write(f"#errors {{ color: #F88; }}\n")
        self.write(f"pre {{ overflow-x: auto; }}\n")
        self.write(f".code {{ background-color:rgba(128,128,128,0.25); margin:1em; padding: 1em; font-size:110%; }}\n")
        self.write(f"</style>\n")
        self.write(f"</head>\n")
        self.write(f"<body>\n")
        self.write(f"<h1>{encode(s.name)}</h1>\n")
        self.write(f"<ul id='errors'></ul>\n")
        for type in s.types:
            self.write_type_ui(type)
        for func in s.functions:
            self.write_function_ui(func)
        self.write(f"<script type='module'>\n")
        self.write(s.get_code("js", self.options))
        self.write("const wgslCode = `\n")
        self.write(s.get_code("wgsl", self.options))
        self.write("`;\n")
        self.write("const wgslLines = wgslCode.split('\\n');\n")
        self.write(f"async function main() {{\n")
        self.indent()
        self.write(f"const $errors = document.getElementById('errors');\n")
        self.write(f"const error = (m) => {{ $errors.appendChild(document.createElement('li')).textContent = m; }};\n")
        self.write(f"const adapter = await navigator.gpu.requestAdapter();\n")
        self.write(f"const device = await adapter.requestDevice();\n")
        self.write(f"const wgslModule = device.createShaderModule({{code: wgslCode}});\n")
        self.write(f"const wgslModuleInfo = await wgslModule.getCompilationInfo();\n")
        self.write("for (let m of wgslModuleInfo.messages) {\n")
        self.write("    console.log(m);\n")
        self.write("    const line = wgslLines[m.lineNum - 1];\n")
        self.write("    error(`WGSL: ${m.message} \"${line}\"`);\n")
        self.write("}\n")
        for type in s.types:
            self.write_type_ui_code(type)
        for func in s.functions:
            self.write_function_ui_code(func)
        self.dedent()
        self.write(f"}}\n")
        self.write(f"main();\n")
        self.write(f"</script>\n")
        self.write(f"</body>\n")
        self.write(f"</html>\n")

    def write_type(self, t: typs.Type):
        self.write_type_ui(t)
        self.write_type_ui_code(t)

    def write_type_ui(self, t: typs.Type):
        if isinstance(t, typs.Struct):
            self.write_struct_ui(t)
        else:
            self.write(f"<h2>{encode(t.name)}</h2>\n")
            self.write(f"<code><pre>{encode(str(t))}</pre></code>\n")

    def write_type_ui_code(self, t: typs.Type):
        if isinstance(t, typs.Struct):
            self.write_struct_ui_code(t)
        else:
            pass

    def write_set(self, s: stmts.Set):
        self.write_expr(s.target)
        self.write(" = ")
        self.write_expr(s.value)

    def write_struct(self, s: typs.Struct):
        self.write_struct_ui(s)
        self.write_struct_ui_code(s)

    def write_struct_ui(self, s: typs.Struct):
        fs: list[typs.Field] = s.fields
        sl = s.layout
        n = len(fs)
        anno = self.options.struct_annotations
        enc_name = encode(s.name)
        self.write(f"<h2>{enc_name}</h2>\n")
        self.write(f"<code><pre>{encode(str(s))}</pre></code>\n")

    def write_struct_ui_code(self, s: typs.Struct):
        self.write(f"let test{s.name} = new {s.name}();\n")
        self.write(f"console.log(test{s.name});\n")
        self.write(f"let test{s.name}GPUBuffer = test{s.name}.createGPUBuffer(device);\n")
        self.write(f"console.log(test{s.name}GPUBuffer);\n")

    def write_function(self, f: funcs.Function):
        self.write_function_ui(f)
        self.write_function_ui_code(f)

    def write_function_ui(self, f: funcs.Function):
        enc_name = encode(f.name)
        self.write(f"<div style='display: block;'>\n")
        self.write(f"<h2>{enc_name}</h2>\n")
        for p in f.parameters:
            self.write_input_ui_for_type(p.name, f.name + "_" + p.name, p.resolved_type or p.parameter_type)
        if f.resolved_type is not None and not f.resolved_type.return_type.is_void:
            self.write_input_ui_for_type("RETURN", f.name + "_return", f.resolved_type.return_type)
        for lang in self.out_langs:
            self.write(f"<div class='code' style='max-width:40%;display:inline-block;vertical-align: top;'>\n")
            self.write(f"<h3>{lang}</h3>\n")
            self.write(f"<code><pre>{encode(f.get_code(lang, self.options))}</pre></code>\n")
            self.write(f"</div>\n")
        self.write(f"<code style='max-width:40%;display:inline-block;'><pre>{encode(str(f))}</pre></code>\n")
        self.write(f"</div>\n")

    def write_input_ui_for_type(self, name: str, id: str, t: "Type"): # type: ignore
        enc_name = encode(name)
        self.write(f"<div>\n")
        self.write(f"<label for='{id}'>{enc_name}</label>\n")
        if t.is_scalar:
            self.write(f"<input id='{id}' type='number'>\n")
        elif t.is_vector:
            axis_names = ["x", "y", "z", "w"]
            n_axes = t.num_elements
            for i in range(n_axes):
                axis_name = axis_names[i]
                self.write(f"<label for='{id}{axis_name}'>{axis_name}</label>\n")
        elif t.is_tensor and len(t.shape) == 2:
            self.write(f"<table id='{id}'>\n")
            nr, nc = t.shape
            for r in range(nr):
                self.write(f"<tr>\n")
                for c in range(nc):
                    self.write(f"<td><input id='{id}_{r}_{c}' type='number' value=0></td>\n")
                self.write(f"</tr>\n")
            self.write(f"</table>\n")
        else:
            self.write(f"<input id='{id}' type='text'>\n")
        self.write(f"</div>\n")

    def write_function_ui_code(self, f: funcs.Function):
        stage_and_auto = self.get_func_stage(f)
        if stage_and_auto is None:
            return
        self.writeln(f"try {{")
        self.indent()
        self.writeln(f"let {f.name}GPUTest = new {f.name}GPU(device, wgslModule);")
        self.writeln(f"await {f.name}GPUTest.exec();")
        self.dedent()
        self.writeln(f"}} catch (e) {{")
        self.indent()
        self.writeln(f"console.error(`{f.name}GPUTest`, e);")
        self.writeln(f"error(`{f.name}GPUTest: ${{e}}`);")
        self.dedent()
        self.writeln(f"}}")

    def write_support_node(self, n: "Node"): # type: ignore
        pass

    def get_typed_name(self, t: typs.Type) -> str:
        if isinstance(t, typs.Integer):
            tn = t.name
            if tn == "sbyte":
                return "Int8"
            elif tn == "byte":
                return "UInt8"
            elif tn == "int":
                return "Int32"
            elif tn == "uint":
                return "UInt32"
            elif tn == "long":
                return "BigInt64"
            elif tn == "ulong":
                return "BigUint64"
            else:
                raise ValueError(f"Invalid integer type: {tn}")
        elif isinstance(t, typs.Float):
            tn = t.name
            if tn == "half":
                return "Float16"
            elif tn == "float":
                return "Float32"
            elif tn == "double":
                return "Float64"
            else:
                raise ValueError(f"Invalid float type: {tn}")
        elif isinstance(t, typs.Vector):
            return self.get_typed_array_name(t.element_type)
        else:
            return t.name
        
    def get_typed_array_name(self, t: typs.Type) -> str:
        return f"{self.get_typed_name(t)}Array"

    def get_js_name(self, t: typs.Type) -> str:
        if isinstance(t, typs.Integer):
            return "number"
        elif isinstance(t, typs.Float):
            return "number"
        elif isinstance(t, typs.Vector):
            return self.get_typed_array_name(t.element_type)
        elif isinstance(t, typs.Array):
            raise NotImplementedError("Arrays not implemented")
        else:
            return t.name

class HTMLLanguage(Language):
    def __init__(self):
        super().__init__("html")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> HTMLWriter: # type: ignore
        return HTMLWriter(out, options)

html_lang = HTMLLanguage()
register_language(html_lang)



from alang.typs import array, struct

def test_simple_struct():
    # https://www.w3.org/TR/WGSL/#structure-member-layout
    s_a = struct(
        "A",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    assert s_a.js_code.strip() == """
class A {
    constructor(buffer, byteOffset, byteLength) {
        byteOffset = byteOffset || 0;
        byteLength = byteLength || 24;
        if (byteLength < 24) throw new Error(`Buffer too small. "A" requires at least 24 bytes, got ${byteLength}`);
        if (buffer instanceof ArrayBuffer) {
            this.buffer = buffer;
        } else {
            this.buffer = new ArrayBuffer(byteLength);
            byteOffset = 0;
        }
        if (byteOffset + byteLength > this.buffer.byteLength) throw new Error(`Buffer overflow. "A" requires ${byteLength} bytes starting at ${byteOffset}, but the buffer is only ${this.buffer.byteLength} bytes long`);
        this.view = new DataView(this.buffer, byteOffset, byteLength);
        this.byteLength = byteLength;
        this.gpuBuffer = null;
        this.isDirty = false;
        this.dirtyBegin = 0;
        this.dirtyEnd = 0;
        this.wArray = new Float32Array(this.buffer, byteOffset + 8, 2);
    }
    dirty(begin, end) {
        if (begin === undefined || end === undefined) { begin = 0; end = this.byteLength; }
        if (this.isDirty) { this.dirtyBegin = Math.min(this.dirtyBegin, begin); this.dirtyEnd = Math.max(this.dirtyEnd, end); }
        else { this.dirtyBegin = begin; this.dirtyEnd = end; }
        this.isDirty = true;
    }
    createGPUBuffer(device, usage) {
        usage = usage || (GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_DST);
        this.gpuBuffer = device.createBuffer({ size: Math.max(this.byteLength, 256), usage: usage, label: "A", mappedAtCreation: false });
        device.queue.writeBuffer(this.gpuBuffer, 0, this.buffer, this.view.byteOffset, this.byteLength);
        this.isDirty = false;
        return this.gpuBuffer;
    }
    get u() { return this.view.getFloat32(0); }
    set u(value) { return this.view.setFloat32(0, value); this.dirty(0, 4); }
    get v() { return this.view.getFloat32(4); }
    set v(value) { return this.view.setFloat32(4, value); this.dirty(4, 8); }
    get w() { return this.wArray; }
    set w(value) { this.wArray.set(value); this.dirty(8, 16); }
    get x() { return this.view.getFloat32(16); }
    set x(value) { return this.view.setFloat32(16, value); this.dirty(16, 20); }
}
""".strip()



from typing import Any, Optional, TextIO, Union

from alang.langs.writer import CodeWriter

class Language:
    def __init__(self, name: str):
        self.name = name
    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> CodeWriter: # type: ignore
        return CodeWriter(out, options)
    def parse_expr(self, code: str):
        raise NotImplementedError()

language_registry = {}
def register_language(lang: Language):
    language_registry[lang.name] = lang

def get_language(language: Optional[Any]) -> Language:
    if language is None:
        return language_registry["a"]
    elif isinstance(language, Language):
        return language
    elif isinstance(language, str):
        if language in language_registry:
            return language_registry[language]
    raise ValueError(f"Invalid language: {language}")



import io
from typing import Any, Callable, Optional, TextIO, TypeVar, Union

import alang.langs as langs

TypeRef = str

Code = str
class CodeOptions:
    def __init__(self, standalone: bool = False, struct_annotations: bool = False, auto_entry_points: bool = False) -> None:
        self.standalone = standalone
        self.struct_annotations = struct_annotations
        self.auto_entry_points = auto_entry_points
        self.entry_points = []

class NodeType:
    ADDRESS = 'address'
    ALIAS = 'alias'
    ARRAY = 'array'
    BINOP = 'binop'
    CONSTANT = 'constant'
    EXPR_STMT = 'expr_stmt'
    FIELD = 'field'
    FLOAT = 'float'
    FUNCALL = 'funcall'
    FUNCTION = 'function'
    FUNCTION_TYPE = 'function_type'
    INDEX = 'index'
    INTEGER = 'integer'
    LOOP = 'loop'
    MEMBER = 'member'
    MODULE = 'module'
    MODULE_TYPE = 'module_type'
    NAME = 'name'
    PARAMETER = 'parameter'
    POINTER = 'pointer'
    RETURN = 'return'
    SET = 'set'
    STRUCT = 'struct'
    SWIZZLE = 'swizzle'
    TENSOR = 'tensor'
    TYPE_NAME = 'type_name'
    VARIABLE = 'variable'
    VOID = 'void'
    VECTOR = 'vector'

next_node_id = 1

class Node:
    def __init__(self, type: str):
        global next_node_id
        self.id = next_node_id
        next_node_id += 1
        self.node_type = type
        self.attributes: dict[str, "NodeAttr"] = {}
        self.links: list[tuple[str, "Node"]] = []
        self.last_backlink: Optional["Node"] = None
        self.resolved_type = None # all nodes get typed
        self.resolved_node = None # some nodes (name, type_name) get resolved to a node
    def append_backlink(self, backlink: "Node", rel: str):
        self.last_backlink = backlink
    def get_rels(self, rel: str) -> list["Node"]:
        return [child for (crel, child) in self.links if crel == rel]
    def get_rel(self, rel: str) -> Optional["Node"]:
        cs = self.get_rels(rel)
        return cs[0] if len(cs) > 0 else None
    def link(self, child: "Node", rel: str) -> "Node":
        if child is None:
            raise ValueError(f"Cannot link {repr(self.node_type)} node to None (rel={rel})")
        if not isinstance(child, Node):
            raise ValueError(f"Cannot link {repr(self.node_type)} node to {repr(child)} (rel={rel})")
        self.links.append((rel, child))
        child.append_backlink(self, rel)
        return self
    def find_reachable_with_type(self, node_type: NodeType) -> list["Node"]:
        reachable = []
        if self.node_type == node_type:
            reachable.append(self)
        for rel, link in self.links:
            reachable.extend(link.find_reachable_with_type(node_type))
        return reachable
    def write_node(self, out, depth, rel):
        if depth > 5:
            out.write("...")
            return
        indent = "    " * depth
        out.write(f"{indent}({self.node_type} (")
        head = ""
        for k, a in self.attributes.items():
            out.write(f"{head}{a.name}={repr(getattr(self, a.private_name))}")
            head = " "
        if len(self.links) > 0:
            out.write(")\n")
            for crel, child in self.links:
                child.write_node(out, depth + 1, crel)
            out.write(f"{indent})\n")
        else:
            out.write("))\n")
    def __str__(self):
        out = io.StringIO()
        self.write_node(out, 0, None)
        return out.getvalue()
    def __repr__(self):
        return str(self)
    def lookup_variable(self, name: str) -> Optional["Variable"]:
        p = self.last_backlink
        while p is not None:
            if isinstance(p, Node):
                return p.lookup_variable(name)
            p = p.parent
        return None
    def resolve_type(self, diags) -> Optional["Type"]: # type: ignore
        raise NotImplementedError(f"resolve_type not implemented for {self.node_type}")
    def get_support_definitions(self, defs: "compiler.SupportDefinitions"): # type: ignore
        pass
    def write_code(self, out: Union[str, TextIO], language: Optional[Any] = None, options: Optional[CodeOptions] = None):
        from alang.compiler import Compiler
        options = options or CodeOptions()
        language = langs.get_language(language)
        compiler = Compiler(self, options)
        compiler.compile()
        options.entry_points = compiler.entry_points
        with language.open_writer(out, options) as writer:
            for s in compiler.support_definitions:
                writer.write_support_node(s)
            writer.write_node(self)
            writer.write_diags(compiler.diags.messages)
    def get_code(self, language: Optional[Any] = None, options: Optional[CodeOptions] = None) -> str:
        out = io.StringIO()
        self.write_code(out, language, options)
        code = out.getvalue()
        return code
    @property
    def code(self) -> str:
        return self.get_code("a")
    @property
    def c_code(self) -> str:
        return self.get_code("c")
    @property
    def html_code(self) -> str:
        return self.get_code("html")
    @property
    def js_code(self) -> str:
        return self.get_code("js")
    @property
    def wgsl_code(self) -> str:
        return self.get_code("wgsl")
    
class NodeAttr:
    def __init__(self, default_value=None):
        self.default_value = default_value
    def __set_name__(self, owner: Node, name: str):
        self.name = name
        self.private_name = f"_{name}"
    def __get__(self, obj: Node, objtype=None):
        if hasattr(obj, self.private_name):
            return getattr(obj, self.private_name)
        return self.default_value
    def __set__(self, obj: Node, value):
        obj.attributes[self.name] = self
        setattr(obj, self.private_name, value)

class NodeLinks:
    def __init__(self):
        self.rel = None
    def __set_name__(self, owner: Node, name: str):
        self.rel = name
    def __get__(self, obj: Node, objtype=None) -> list[Node]:
        return obj.get_rels(self.rel)
    def __set__(self, obj: Node, value: list[Node]):
        i = 0
        while i < len(obj.links):
            if obj.links[i][0] == self.rel:
                ch = obj.links.pop(i)[1]
                ch.last_backlink = None
            else:
                i += 1
        for child in value:
            obj.link(child, self.rel)

class NodeLink:
    def __init__(self):
        self.rel = None
    def __set_name__(self, owner: Node, name: str):
        self.rel = name
    def __get__(self, obj: Node, objtype=None) -> list[Node]:
        return obj.get_rel(self.rel)
    def __set__(self, obj: Node, value: Optional[Node]):
        i = 0
        while i < len(obj.links):
            if obj.links[i][0] == self.rel:
                ch = obj.links.pop(i)[1]
                ch.last_backlink = None
            else:
                i += 1
        if value is not None:
            obj.link(value, self.rel)

class Visitor:
    def visit(self, node: Node, parent: Node, rel: str, acc):
        raise NotImplementedError()
    def visit_node(self, node: Node, parent: Node, rel: str, acc):
        if node.node_type == NodeType.ALIAS:
            return self.visit_alias(node, parent, rel, acc)
        elif node.node_type == NodeType.ARRAY:
            return self.visit_array(node, parent, rel, acc)
        elif node.node_type == NodeType.BINOP:
            return self.visit_binop(node, parent, rel, acc)
        elif node.node_type == NodeType.CONSTANT:
            return self.visit_constant(node, parent, rel, acc)
        elif node.node_type == NodeType.EXPR_STMT:
            return self.visit_expr_stmt(node, parent, rel, acc)
        elif node.node_type == NodeType.FIELD:
            return self.visit_field(node, parent, rel, acc)
        elif node.node_type == NodeType.FLOAT:
            return self.visit_float(node, parent, rel, acc)
        elif node.node_type == NodeType.FUNCALL:
            return self.visit_funcall(node, parent, rel, acc)
        elif node.node_type == NodeType.FUNCTION:
            return self.visit_function(node, parent, rel, acc)
        elif node.node_type == NodeType.INDEX:
            return self.visit_index(node, parent, rel, acc)
        elif node.node_type == NodeType.INTEGER:
            return self.visit_integer(node, parent, rel, acc)
        elif node.node_type == NodeType.LOOP:
            return self.visit_loop(node, parent, rel, acc)
        elif node.node_type == NodeType.MEMBER:
            return self.visit_member(node, parent, rel, acc)
        elif node.node_type == NodeType.MODULE:
            return self.visit_module(node, parent, rel, acc)
        elif node.node_type == NodeType.NAME:
            return self.visit_name(node, parent, rel, acc)
        elif node.node_type == NodeType.PARAMETER:
            return self.visit_parameter(node, parent, rel, acc)
        elif node.node_type == NodeType.POINTER:
            return self.visit_pointer(node, parent, rel, acc)
        elif node.node_type == NodeType.RETURN:
            return self.visit_return(node, parent, rel, acc)
        elif node.node_type == NodeType.SET:
            return self.visit_set(node, parent, rel, acc)
        elif node.node_type == NodeType.STRUCT:
            return self.visit_struct(node, parent, rel, acc)
        elif node.node_type == NodeType.SWIZZLE:
            return self.visit_swizzle(node, parent, rel, acc)
        elif node.node_type == NodeType.TENSOR:
            return self.visit_tensor(node, parent, rel, acc)
        elif node.node_type == NodeType.VARIABLE:
            return self.visit_variable(node, parent, rel, acc)
        elif node.node_type == NodeType.VECTOR:
            return self.visit_vector(node, parent, rel, acc)
        elif node.node_type == NodeType.VOID:
            return self.visit_void(node, parent, rel, acc)
        else:
            missing_code = f"elif node.node_type == NodeType.{node.node_type.upper()}:\n    return self.visit_{node.node_type.lower()}(node, parent, rel, acc)"
            print(missing_code)
            return acc
    def visit_address(self, node: "Address", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_alias(self, node: "Alias", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_array(self, node: "Array", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_binop(self, node: "Binop", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_constant(self, node: "Constant", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_expr_stmt(self, node: "ExprStmt", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_field(self, node: "Field", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_float(self, node: "Float", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_loop(self, node: "Loop", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_funcall(self, node: "Funcall", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_function(self, node: "Function", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_index(self, node: "Index", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_integer(self, node: "Integer", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_member(self, node: "Member", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_module(self, node: "Module", parent: Node, rel: str, acc):
        return acc
    def visit_name(self, node: "Name", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_parameter(self, node: "Parameter", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_pointer(self, node: "Pointer", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_return(self, node: "Return", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_set(self, node: "Set", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_swizzle(self, node: "Swizzle", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_struct(self, node: "Struct", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_tensor(self, node: "Tensor", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_variable(self, node: "Variable", parent: Node, rel: str, acc):
        return acc
    def visit_vector(self, node: "Vector", parent: Node, rel: str, acc): # type: ignore
        return acc
    def visit_void(self, node: "Void", parent: Node, rel: str, acc): # type: ignore
        return acc
    
class DepthFirstVisitor(Visitor):
    def visit(self, node: Node, parent: Node, rel: str, acc):
        sacc = []
        for crel, child in node.links:
            sacc.append((crel, self.visit(child, node, crel, acc)))
        return self.visit_node(node, parent, rel, sacc)

class BreadthFirstVisitor(Visitor):
    def visit(self, node: Node, parent: Node, rel: str, acc):
        cacc = self.visit_node(node, parent, rel, acc)
        for crel, child in node.links:
            self.visit(child, node, crel, cacc)

class Expression(Node):
    def __init__(self, node_type: NodeType):
        super().__init__(node_type)
    def __add__(self, other: "Expression") -> "Expression":
        return self.add(other)
    def add(self, other: "Expression") -> "Expression":
        from alang.exprs import Binop
        return Binop(self, "+", other)
    def __sub__(self, other: "Expression") -> "Expression":
        return self.sub(other)
    def sub(self, other: "Expression") -> "Expression":
        from alang.exprs import Binop
        return Binop(self, "-", other)
    def __mul__(self, other: "Expression") -> "Expression":
        return self.mul(other)
    def mul(self, other: "Expression") -> "Expression":
        from alang.exprs import Binop
        return Binop(self, "*", other)
    def __truediv__(self, other: "Expression") -> "Expression":
        return self.div(other)
    def div(self, other: "Expression") -> "Expression":
        from alang.exprs import Binop
        return Binop(self, "/", other)

class Statement(Node):
    def __init__(self, node_type: NodeType):
        super().__init__(node_type)

class Block(Node):
    types = NodeLinks()
    variables = NodeLinks()
    functions: list["funcs.Functions"] = NodeLinks() # type: ignore
    statements = NodeLinks()
    def __init__(self, type: NodeType, can_define_types: bool, can_define_functions: bool, can_define_variables: bool, can_define_statements: bool):
        super().__init__(type)
        self.can_define_types = can_define_types
        self.can_define_functions = can_define_functions
        self.can_define_variables = can_define_variables
        self.can_define_statements = can_define_statements
    def parse_type(self, expr: Optional[Code]) -> Optional["Type"]: # type: ignore
        from alang.typs import try_resolve_type
        return try_resolve_type(expr, self)
    def parse_expr(self, expr: Optional[Code]) -> Optional["Expression"]:
        from alang.exprs import parse_expr
        return parse_expr(expr, self)
    def parse_stmt(self, stmt: Optional[Code]) -> Optional["Expression"]:
        return self.parse_expr(stmt)
    def append_any(self, child: Node):
        if child.node_type == NodeType.ARRAY:
            self.link(child, "types")
        elif child.node_type == NodeType.STRUCT:
            self.link(child, "types")
        elif child.node_type == NodeType.VARIABLE:
            self.link(child, "variables")
        elif child.node_type == NodeType.FUNCTION:
            self.link(child, "functions")
        elif isinstance(child, Statement):
            self.link(child, "statements")
        else:            
            raise ValueError(f"Cannot append {child.node_type} to {self.node_type}")
    def array(self, element_type: str, length: Optional[int] = None) -> "Array": # type: ignore
        if self.can_define_types:
            from alang.typs import Array
            a = Array(element_type, length)
            self.link(a, "types")
            return a
        else:
            raise ValueError(f"Cannot define array in {self.node_type}")
    def call(self, func: str, *args: list) -> "Block":
        from alang.exprs import Funcall
        from alang.stmts import ExprStmt
        func = self.parse_expr(func)
        args = [self.parse_expr(x) for x in args]
        funcall = Funcall(func, args)
        call_stmt = ExprStmt(funcall)
        self.stmt(call_stmt)
        return self
    def define(self, name: str, *parameters: list) -> "Function": # type: ignore
        if self.can_define_functions:
            from alang.funcs import Function
            f = Function(name, None, *parameters)
            self.link(f, "functions")
            return f
        else:
            raise ValueError(f"Cannot define function in {self.node_type}")
    def loop(self, var: str, count: Code, *statements: list[Code]) -> "Block": # type: ignore
        from alang.stmts import Loop
        count = self.parse_expr(count)
        loop = Loop(var, count, *statements)
        self.stmt(loop)
        return self
    def ret(self, value: Optional[Code]) -> "Block":
        from alang.stmts import Return
        r = Return(self.parse_expr(value))
        self.stmt(r)
        return self
    def set(self, lhs: Code, rhs: Code) -> "Block":
        lhs = self.parse_expr(lhs)
        rhs = self.parse_expr(rhs)
        v = self.lookup_variable(lhs)
        if v is None:
            if self.can_define_variables:
                v = Variable(lhs)
                self.link(v, "variables")
            else:
                raise ValueError(f"Variable {lhs} not defined")
        else:
            raise ValueError(f"Variable {lhs} already defined")
        return self
    def stmt(self, stmt):
        if not self.can_define_statements:
            raise ValueError(f"Cannot define return statements in {self.node_type}")
        self.link(stmt, "statements")
    def struct(self, name: str, *fields: list) -> "Struct": # type: ignore
        if self.can_define_types:
            from alang.typs import Struct
            s = Struct(name)
            for field in fields:
                s.field(*field)
            self.link(s, "types")
            return s
        else:
            raise ValueError(f"Cannot define struct in {self.node_type}")
    def get_default_address_space(self) -> Optional[str]:
        return None
    def var(
            self,
            name: str, type: Optional["Type"] = None, initial_value: Expression = None, # type: ignore
            bind_group: Optional[int] = None, binding: Optional[int] = None,
            address_space: Optional[str] = None, access_mode: Optional[str] = None,
            ) -> "Block":
        if self.can_define_variables:
            type = self.parse_type(type)
            if address_space is None:
                address_space = self.get_default_address_space()
            v = Variable(name,
                         type=type, initial_value=initial_value,
                         bind_group=bind_group, binding=binding,
                         address_space=address_space, access_mode=access_mode)
            self.link(v, "variables")
            return self
        else:
            raise ValueError(f"Variable {name} not defined")
        
class AddressSpace:
    FUNCTION = 'function'
    PRIVATE = 'private'
    STORAGE = 'storage'
    UNIFORM = 'uniform'
    WORKGROUP = 'workgroup'

class AccessMode:
    READ = 'read'
    WRITE = 'write'
    READ_WRITE = 'read_write'

class Module(Block):
    name = NodeAttr()

    def __init__(self, name: str, can_define_statements: bool = False):
        super().__init__(NodeType.MODULE, can_define_types=True, can_define_functions=True, can_define_variables=True, can_define_statements=can_define_statements)
        if name is not None:
            self.name = name

    def get_default_address_space(self) -> Optional[str]:
        return AddressSpace.PRIVATE

    def resolve_type(self, diags: "compiler.Diagnostics") -> "typs.Type": # type: ignore
        from alang.typs import ModuleType
        return ModuleType(self.name)

def get_default_access_mode_for_address_space(address_space: str) -> str:
    if address_space == AddressSpace.FUNCTION:
        return AccessMode.READ_WRITE
    elif address_space == AddressSpace.PRIVATE:
        return AccessMode.READ_WRITE
    elif address_space == AddressSpace.WORKGROUP:
        return AccessMode.READ_WRITE
    elif address_space == AddressSpace.UNIFORM:
        return AccessMode.READ
    elif address_space == AddressSpace.STORAGE:
        return AccessMode.READ
    else:
        raise ValueError(f"Invalid address space: {address_space}")

class Variable(Node):
    # https://www.w3.org/TR/WGSL/#var-decls
    name = NodeAttr()
    variable_type = NodeLink()
    initial_value = NodeLink()
    address_space = NodeAttr()
    access_mode = NodeAttr()

    def __init__(
            self,
            name: str, type: Optional["Type"] = None, initial_value: Expression = None, # type: ignore
            bind_group: Optional[int] = None, binding: Optional[int] = None,
            address_space: Optional[str] = None, access_mode: Optional[str] = None,
            ):
        super().__init__(NodeType.VARIABLE)
        self.name = name
        self.variable_type = type
        self.initial_value = initial_value
        self.bind_group = bind_group
        self.binding = binding
        self.address_space = address_space
        self.access_mode = access_mode
        # The address space must be specified if the access mode is specified.
        if self.access_mode is not None and self.address_space is None:
            raise ValueError(f"The address space must be specified if the access mode is specified")
        # The access mode always has a default value, and except for variables in the storage address space, must not be specified in the WGSL source.
        if self.access_mode is not None and self.address_space != AddressSpace.STORAGE:
            raise ValueError(f"Cannot specify access mode for {repr(self.address_space)} address space")
        if self.address_space is not None and self.access_mode is None:
            self.access_mode = get_default_access_mode_for_address_space(self.address_space)
    
    def resolve_type(self, diags: "compiler.Diagnostics") -> "typs.Type": # type: ignore
        if self.variable_type is None:
            return None
        return self.variable_type.resolved_type

global_module = Module("global", can_define_statements=True)

def define(name: str, *parameters: list) -> "Function": # type: ignore
    return global_module.define(name, *parameters)

def loop(var: str, count, body = None) -> "Array": # type: ignore
    return global_module.loop(var, count, body)




from alang.nodes import define
from alang.typs import struct, tensor_type, Tensor, float_type, int_type

from test_html import write_standalone_html

def test_tensor_from_element_type():
    t = tensor_type((3, 5, 7, 11), int_type)
    assert t.element_type.name == "int"

def test_tensor_from_str_element_type():
    t = tensor_type((3, 5, 7, 11), "short")
    assert t.element_type.name == "short"

def test_tensor_from_str_element_type():
    s = struct("S", ("t", "byte42x69x7"))
    t: Tensor = s.fields[0].field_type
    assert t.node_type == "tensor"
    assert t.element_type.name == "byte"

def test_tensor_wgsl():
    t = tensor_type((3, 5, 7, 11), float_type)
    s = struct("StructWithTensor", ("t", t))
    assert s.wgsl_code.strip() == """
alias float3x5x7x11 = array<f32, 1155>;
struct StructWithTensor {
    t: float3x5x7x11
}""".strip()

def test_3x5_times_5x7_type():
    t1 = tensor_type((3, 5), int_type)
    t2 = tensor_type((5, 7), int_type)
    t3 = t1 @ t2
    assert t3.shape == (3, 7)

def test_1x2_matmul_2x1():
    at = tensor_type((1, 2), int_type)
    bt = tensor_type((2, 1), int_type)
    ot = at @ bt
    f = define("f").param("a", at).param("b", bt).ret("a @ b")
    write_standalone_html("test_1x2_times_2x1", f)
    code = f.wgsl_code
    assert code.strip() == """
alias int1x2 = array<i32, 2>;
alias int2x1 = array<i32, 2>;
fn mul_int1x2_int2x1(a: int1x2, b: int2x1) -> int1x1 {
    var o: int1x1;
    for (var out_r: i32 = 0; out_r < 1; out_r++) {
        for (var out_c: i32 = 0; out_c < 1; out_c++) {
            o[((out_r * 1) + out_c)] = ((a[(out_r * 2)] * b[out_c]) + (a[((out_r * 2) + 1)] * b[(1 + out_c)]));
        }
    }
    return o;
}
alias int1x1 = array<i32, 1>;
fn f(a: int1x2, b: int2x1) -> int1x1 {
    return mul_int1x2_int2x1(a, b);
}""".strip()

def test_3x5_matmul_5x7():
    t1 = tensor_type((3, 5), int_type)
    t2 = tensor_type((5, 7), int_type)
    t3 = t1 @ t2
    f = define("f").param("a", t1).param("b", t2).ret("a @ b")
    write_standalone_html("test_3x5_times_5x7", f)
    code = f.wgsl_code
    assert code.strip() == """
alias int3x5 = array<i32, 15>;
alias int5x7 = array<i32, 35>;
fn mul_int3x5_int5x7(a: int3x5, b: int5x7) -> int3x7 {
    var o: int3x7;
    for (var out_r: i32 = 0; out_r < 3; out_r++) {
        for (var out_c: i32 = 0; out_c < 7; out_c++) {
            o[((out_r * 7) + out_c)] = (((((a[(out_r * 5)] * b[out_c]) + (a[((out_r * 5) + 1)] * b[(7 + out_c)])) + (a[((out_r * 5) + 2)] * b[(14 + out_c)])) + (a[((out_r * 5) + 3)] * b[(21 + out_c)])) + (a[((out_r * 5) + 4)] * b[(28 + out_c)]));
        }
    }
    return o;
}
alias int3x7 = array<i32, 21>;
fn f(a: int3x5, b: int5x7) -> int3x7 {
    return mul_int3x5_int5x7(a, b);
}""".strip()

def test_standalone_tensor():
    t = tensor_type((3, 5, 7, 11), float_type)
    s = struct("StructWithTensor", ("t", t))
    html = write_standalone_html("test_standalone_tensor", s)
    assert html.index("Float32Array") > 0

def test_const_flat_index():
    t = tensor_type((3, 5, 7, 11), float_type)
    def tt(index, expected):
        fi = t.get_flat_index(index)
        assert fi.node_type == "constant"
        assert fi.value == expected
    tt((0, 0, 0, 0), 0)
    tt((0, 0, 0, 1), 1)
    tt((1, 0, 0, 0), 1 * 5 * 7 * 11)
    tt((1, 2, 3, 4), 1 * 5 * 7 * 11 + 2 * 7 * 11 + 3 * 11 + 4)
    tt((2, 4, 6, 10),  2 * 5 * 7 * 11 + 4 * 7 * 11 + 6 * 11 + 10)

def test_const_flat_index():
    t = tensor_type((3, 5, 7, 11), float_type)
    def tt(index, expected):
        fi = t.get_flat_index(index)
        fi_code: str = fi.wgsl_code
        e_index = fi_code.find("//")
        if e_index > 0:
            fi_code = fi_code[:e_index].strip()
        assert fi_code == expected
    tt((0, 0, 0, 0), "0")
    tt((0, 0, 0, 1), "1")
    tt((0, 0, 0, "x"), "x")
    tt(("y", 0, 0, 0), "(y * 385)")
    tt(("y", 0, 0, "x"), "((y * 385) + x)")



from typing import Optional, TextIO, Union

import alang.nodes as nodes

class CodeWriter:
    def __init__(self, path_or_io: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        self.options = options
        self.indent_level = 0
        self.needs_indent = True
        if self.options is None:
            from nodes import CodeOptions
            self.options = CodeOptions()
        if type(path_or_io) is str:
            self.out = open(path_or_io, "w")
            self.owner = True
        else:
            self.out = path_or_io
            self.owner = False
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
    def close(self):
        if self.owner and self.out is not None:
            self.out.close()
            self.out = None
            self.owner = False

    def indent(self):
        self.indent_level += 1
    def dedent(self):
        self.indent_level = max(0, self.indent_level - 1)
    def write(self, s: str):
        lines = s.split("\n")
        n_lines = len(lines)
        for i, line in enumerate(lines):
            if len(line) > 0:
                if self.needs_indent:
                    self.out.write("    " * self.indent_level)
                    self.needs_indent = False
                self.out.write(line)
            if i < n_lines - 1:
                self.out.write("\n")
                self.needs_indent = True
    def writeln(self, s: str):
        self.write(s + "\n")

    def error(self, message: str):
        self.write_multiline_comment(f"ERROR! {message}")

    def warning(self, message: str):
        self.write_multiline_comment(f"WARNING: {message}")

    def write_error_value(self, message: str, type: Optional["typs.Type"] = None): # type: ignore
        self.write_zero_value_for_type(type)
        self.write_inline_comment(f"ERROR! {message}")

    def get_func_stage(self, f: "funcs.Function") -> Optional[tuple]: # type: ignore
        for e, e_stage, out_e in self.options.entry_points:
            if e.name == f.name:
                return e_stage, out_e
        return None

    def write_inline_comment(self, comment: str):
        raise NotImplementedError

    def write_line_comment(self, comment: str):
        raise NotImplementedError

    def write_diags(self, diags: list["DiagnosticMessage"]): # type: ignore
        for d in diags:
            m = d.message
            if d.node is not None:
                m += f" {d.node}"
            m = m.strip()
            if d.kind == "error":
                self.error(m)
            elif d.kind == "warning":
                self.warning(m)

    def write_alias(self, b: "Alias"): # type: ignore
        raise NotImplementedError

    def write_array(self, b: "Array"): # type: ignore
        raise NotImplementedError

    def write_binop(self, b: "Binop"): # type: ignore
        raise NotImplementedError

    def write_constant(self, c: "Constant"): # type: ignore
        raise NotImplementedError

    def write_expr(self, e: "Expression"): # type: ignore
        return self.write_node(e)

    def write_expr_stmt(self, e: "ExprStmt"): # type: ignore
        raise NotImplementedError

    def write_funcall(self, f: "Funcall"): # type: ignore
        raise NotImplementedError

    def write_function(self, f):
        raise NotImplementedError

    def write_index(self, i: "Index"): # type: ignore
        raise NotImplementedError

    def write_loop(self, f: "Loop"): # type: ignore
        raise NotImplementedError

    def write_module(self, m: "modules.Module"): # type: ignore
        for type in m.types:
            self.write_type(type)
        for var in m.variables:
            self.write_variable(var)
        for func in m.functions:
            self.write_function(func)

    def write_multiline_comment(self, comment: str):
        lines = comment.split("\n")
        for line in lines:
            self.write_line_comment(line)

    def write_name(self, n: "Name"): # type: ignore
        raise NotImplementedError
    
    def write_node(self, n: "Node"): # type: ignore
        if n.node_type == nodes.NodeType.ALIAS:
            self.write_alias(n)
        elif n.node_type == nodes.NodeType.ARRAY:
            self.write_array(n)
        elif n.node_type == nodes.NodeType.BINOP:
            self.write_binop(n)
        elif n.node_type == nodes.NodeType.EXPR_STMT:
            self.write_expr_stmt(n)
        elif n.node_type == nodes.NodeType.CONSTANT:
            self.write_constant(n)
        elif n.node_type == nodes.NodeType.FUNCALL:
            self.write_funcall(n)
        elif n.node_type == nodes.NodeType.FUNCTION:
            self.write_function(n)
        elif n.node_type == nodes.NodeType.INDEX:
            self.write_index(n)
        elif n.node_type == nodes.NodeType.LOOP:
            self.write_loop(n)
        elif n.node_type == nodes.NodeType.MODULE:
            self.write_module(n)
        elif n.node_type == nodes.NodeType.NAME:
            self.write_name(n)
        elif n.node_type == nodes.NodeType.POINTER:
            self.write_pointer(n)
        elif n.node_type == nodes.NodeType.RETURN:
            self.write_return(n)
        elif n.node_type == nodes.NodeType.SET:
            self.write_set(n)
        elif n.node_type == nodes.NodeType.STRUCT:
            self.write_struct(n)
        elif n.node_type == nodes.NodeType.VARIABLE:
            self.write_variable(n)
        else:
            raise ValueError(f"Cannot write node of type \"{n.node_type}\"")

    def write_pointer(self, r: "Pointer"): # type: ignore
        raise NotImplementedError

    def write_return(self, r: "Return"): # type: ignore
        raise NotImplementedError

    def write_set(self, r: "Set"): # type: ignore
        raise NotImplementedError

    def write_stmt(self, s: "Statement"): # type: ignore
        return self.write_node(s)

    def write_struct(self, s):
        raise NotImplementedError

    def write_support_node(self, n: "Node"): # type: ignore
        return self.write_node(n)

    def write_type(self, t: "typs.Type"): # type: ignore
        self.write_node(t)

    def write_variable(self, v: "Variable"): # type: ignore
        raise NotImplementedError

    def write_zero_value_for_type(self, type: Optional["typs.Type"] = None): # type: ignore
        raise NotImplementedError



"""WGSL https://www.w3.org/TR/WGSL/"""

from typing import Optional, TextIO, Union

from alang.langs.language import Language, register_language
from alang.langs.writer import CodeWriter
import alang.nodes as nodes
import alang.exprs as exprs
import alang.typs as typs
import alang.funcs as funcs
import alang.stmts as stmts

class WGSLWriter(CodeWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)

    def write_alias(self, a: "Alias"): # type: ignore
        self.write(f"alias {a.name} = ")
        self.write_type_ref(a.aliased_type)
        self.write(";\n")

    def write_binop(self, b: exprs.Binop):
        support_name = b.get_support_lib_function_name()
        if support_name is not None:
            self.write(support_name)
            self.write("(")
            self.write_expr(b.left)
            self.write(", ")
            self.write_expr(b.right)
            self.write(")")
        else:
            self.write("(")
            self.write_expr(b.left)
            self.write(" ")
            self.write(b.operator.op)
            self.write(" ")
            self.write_expr(b.right)
            self.write(")")

    def write_block(self, f: stmts.Block):
        self.writeln("{")
        self.indent()
        for v in f.variables:
            self.write_variable(v)
        for s in f.statements:
            self.write_node(s)
        self.dedent()
        self.writeln("}")

    def write_constant(self, c: "Constant"): # type: ignore
        self.write(repr(c.value))

    def write_expr_stmt(self, e: stmts.ExprStmt):
        self.write_expr(e.expression)
        self.write(";\n")

    def write_index(self, i: "Index"): # type: ignore
        self.write_expr(i.base)
        self.write("[")
        for ri, r in enumerate(i.ranges):
            if ri > 0:
                self.write(", ")
            if r is None:
                self.write(":")
            else:
                self.write_expr(r)
        self.write("]")

    def write_loop(self, f: stmts.Loop):
        self.write(f"for (var {f.var}: ")
        self.write_type_ref(typs.int_type)
        self.write(f" = 0; {f.var} < ")
        self.write_expr(f.count)
        self.writeln(f"; {f.var}++) {{")
        self.indent()
        for s in f.statements:
            self.write_node(s)
        self.dedent()
        self.writeln("}")

    def write_funcall(self, c: exprs.Funcall):
        f = c.func
        if f is None:
            self.write_error_value("Missing function")
            return
        if f.node_type == nodes.NodeType.FUNCTION:
            self.write(f.name)
        else:
            self.write_expr(f)
        self.write("(")
        for i, arg in enumerate(c.args):
            if i > 0:
                self.write(", ")
            self.write_expr(arg)
        self.write(")")

    def write_function(self, f: funcs.Function):
        stage_and_auto = self.get_func_stage(f)
        if f.stage is not None:
            self.writeln(f"@{f.stage}")
        self.write(f"fn {f.name}(")
        ps = f.parameters
        for i, param in enumerate(ps):
            if param.location is not None:
                self.write(f"@location({param.location}) ")
            self.write(f"{param.name}: ")
            self.write_type_ref(param.parameter_type)
            if i < len(ps) - 1:
                self.write(", ")
        self.write(") ")
        if not f.return_type.is_void:
            self.write("-> ")
            self.write_type_ref(f.return_type)
            self.write(" ")
        self.write_block(f)
        if stage_and_auto is not None:
            self.write_auto_entry_point_function(f, *stage_and_auto)

    def write_auto_entry_point_function(self, original_f: funcs.Function, stage: str, f: funcs.Function):
        pts = original_f.resolved_type.parameter_types
        for i, p in enumerate(original_f.parameters):
            self.write(f"@group(0) @binding({i}) var<storage> {p.name}_auto_compute: ")
            self.write_type_ref(pts[i])
            self.write(";\n")
        workgroup_size = f.workgroup_size or 1
        self.writeln(f"@{stage} @workgroup_size({workgroup_size})")
        self.writeln(f"fn {f.name}() {{")
        self.indent()
        self.write(f"{original_f.name}(")
        for i, p in enumerate(original_f.parameters):
            if i > 0:
                self.write(", ")
            self.write(p.name + "_auto_compute")
        self.writeln(");")
        self.dedent()
        self.writeln("}")

    def write_line_comment(self, comment: str):
        self.write("// ")
        self.write(comment)
        self.write("\n")

    def write_name(self, n: exprs.Name):
        self.write(n.name)

    def write_return(self, r: stmts.Return):
        self.write("return")
        if r.value is not None:
            self.write(" ")
            self.write_expr(r.value)
        self.write(";\n")

    def write_set(self, s: stmts.Set):
        self.write_expr(s.target)
        self.write(" = ")
        self.write_expr(s.value)
        self.write(";\n")

    def write_struct(self, s: typs.Struct):
        fs = s.fields
        sl = s.layout
        n = len(fs)
        anno_col = 46
        anno = self.options.struct_annotations
        self.write(f"struct {s.name} {{")
        def write_anno(o, a, s, text_len):
            if anno:
                tab = " " * max(anno_col - text_len, 1)
                o_text = ""
                if o is not None:
                    o_text = f"offset({o})"
                a_text = f"align({a})"
                self.write(f"{tab}// {o_text.ljust(12)}{a_text.ljust(10)}size({s})")
        write_anno(None, sl.align, sl.byte_size, len(s.name) + 8)
        self.write("\n")
        for i, field in enumerate(fs):
            field_type = self.get_type_name(field.field_type)
            self.write(f"    {field.name}: {field_type}")
            text_len = len(field.name) + len(field_type) + 5
            if i < n - 1:
                self.write(",")
                text_len += 1
            fl = sl.fields[i]
            write_anno(fl.offset, fl.align, fl.byte_size, text_len)
            self.write("\n")
        self.write("}\n")

    def write_type_ref(self, t: typs.Type):
        self.write(self.get_type_name(t))

    def write_variable(self, v: nodes.Variable):
        self.write("var")
        if v.address_space is not None and v.address_space != "function":
            self.write(f"<{v.address_space}")
            if v.access_mode is not None and v.address_space == "storage":
                self.write(", ")
                self.write(v.access_mode)
            self.write("> ")
        else:
            self.write(" ")
        self.write(v.name)
        self.write(": ")
        self.write_type_ref(v.resolved_type or v.variable_type)
        self.write(";\n")

    def get_type_name(self, t: typs.Type) -> str:
        if t is None:
            return "void"
        elif isinstance(t, typs.Array):
            n = t.num_elements
            element_type = self.get_type_name(t.element_type)
            return f"array<{element_type}, {n}>" if n is not None else f"array<{element_type}>"
        elif isinstance(t, typs.Integer):
            tn = t.name
            if tn == "sbyte":
                return "i8"
            elif tn == "byte":
                return "u8"
            elif tn == "int":
                return "i32"
            elif tn == "uint":
                return "u32"
            elif tn == "long":
                return "i64"
            elif tn == "ulong":
                return "u64"
            else:
                raise ValueError(f"Invalid integer type: {tn}")
        elif isinstance(t, typs.Float):
            tn = t.name
            if tn == "half":
                return "f16"
            elif tn == "float":
                return "f32"
            elif tn == "double":
                return "f64"
            else:
                raise ValueError(f"Invalid float type: {tn}")
        elif isinstance(t, typs.Vector):
            n = t.size
            if n == 2:
                vec_type = "vec2"
            elif n == 3:
                vec_type = "vec3"
            elif n == 4:
                vec_type = "vec4"
            else:
                raise ValueError(f"Invalid vector size: {n}")
            element_type = self.get_type_name(t.element_type)
            return f"{vec_type}<{element_type}>"
        elif isinstance(t, typs.Pointer):
            element_type = self.get_type_name(t.element_type)
            address_space = t.address_space
            if address_space is None:
                address_space = "function"
            if t.access_mode is not None:
                return f"ptr<{address_space}, {element_type}, {t.access_mode}>"
            else:
                return f"ptr<{address_space}, {element_type}>"
        else:
            return t.name

class WGSLLanguage(Language):
    def __init__(self):
        super().__init__("wgsl")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> WGSLWriter: # type: ignore
        return WGSLWriter(out, options)

wgsl_lang = WGSLLanguage()
register_language(wgsl_lang)




wrote_vec_def = set()
wrote_mat_def = set()
wrote_mm = set()

def get_mat_type(shape) -> str:
    r, c = shape
    if c == 1 or r == 1:
        n = max(r, c)
        if n == 1:
            return "f32"
        elif n <= 4:
            return f"vec{n}<f32>"
        else:
            return f"vec{n}f"
    else:
        if r <= 4 and c <= 4:
            return f"mat{r}x{c}<f32>"
        else:
            return f"mat{r}x{c}f"

def get_mm_name(a_shape, b_shape, out):
    global wrote_mm
    a_r, a_c = a_shape
    b_r, b_c = b_shape
    if a_c != b_r:
        raise ValueError(f"Matrices can't be multiplied: {a_shape} x {b_shape}")
    a_type = get_mat_type(a_shape).replace("<f32>", "f")
    b_type = get_mat_type(b_shape).replace("<f32>", "f")
    name = f"mul_{a_type}_{b_type}"
    if name not in wrote_mm:
        write_mm_func(name, a_shape, b_shape, out)
    return name

def write_vec_defs(r, out):
    global wrote_vec_def
    if r <= 4 or r in wrote_vec_def:
        return
    wrote_vec_def.add(r)
    vec_type = get_mat_type((r, 1))
    num_vec4s = r // 4
    rem_vec_len = r % 4
    # Data Type
    out.write(f"struct {vec_type} {{\n")
    for i in range(num_vec4s):
        out.write(f"    row{i*4}: vec4<f32>")
        if i < num_vec4s - 1 or rem_vec_len > 0:
            out.write(",\n")
        else:
            out.write("\n")
    if rem_vec_len > 0:
        out.write(f"    row{num_vec4s*4}: vec{rem_vec_len}<f32>;\n")
    out.write("};\n")
    # Dot Product
    out.write(f"fn dot{r}(a: {vec_type}, b: {vec_type}) -> f32 {{\n")
    out.write("    return ")
    for i in range(num_vec4s):
        out.write(f"dot(a.row{i*4}, b.row{i*4})")
        if i < num_vec4s - 1 or rem_vec_len > 0:
            out.write(" + ")
    if rem_vec_len > 0:
        out.write(f"dot(a.row{num_vec4s*4}, b.row{num_vec4s*4})")
    out.write(";\n")
    out.write("}\n")

def write_mat_defs(r, c, out):
    global wrote_mat_def
    if r <= 4 and c <= 4 or (r, c) in wrote_mat_def:
        return
    wrote_mat_def.add((r, c))
    mat_type = get_mat_type((r, c))
    col_type = get_mat_type((r, 1))
    # Data Type
    out.write(f"struct {mat_type} {{\n")
    for col in range(c):
        out.write(f"    col{col}: {col_type}")
        if col < c - 1:
            out.write(",\n")
        else:
            out.write("\n")
    out.write("};\n")

def write_mat_row(name, shape, row, out):
    r, c = shape
    if r == 1 or c == 1:
        out.write(f"{name}")
        return
    row_type = get_mat_type((1, c))
    out.write(f"{row_type}(")
    for col in range(c):
        out.write(f"{name}[{row}, {col}]")
        if col < c - 1:
            out.write(", ")
    out.write(")")

def write_mat_col(name, shape, col, out):
    r, c = shape
    if r == 1 or c == 1:
        out.write(f"{name}")
        return
    col_type = get_mat_type((r, 1))
    if r > 4 or c > 4:
        out.write(f"{name}.col{col}")
        return
    out.write(f"{name}[{col}]")

def write_mm_func(name, a_shape, b_shape, out):
    global wrote_mm
    wrote_mm.add(name)
    a_r, a_c = a_shape
    b_r, b_c = b_shape
    if a_c != b_r:
        raise ValueError(f"Matrices can't be multiplied: {a_shape} x {b_shape}")
    out_r, out_c = a_r, b_c
    a_type = get_mat_type(a_shape)
    b_type = get_mat_type(b_shape)
    out_type = get_mat_type((out_r, out_c))
    out_col_type = get_mat_type((out_r, 1))
    dot_name = "dot"
    if a_c > 4:
        write_vec_defs(a_c, out)
        dot_name = f"dot{a_c}"
    out.write(f"fn {name}(a: {a_type}, b: {b_type}) -> {out_type} {{\n")
    # Calculate the MM one column at a time
    for out_col in range(b_c):
        out.write(f"    let col{out_col}: {out_col_type} = ")
        if out_col_type != "f32":
            out.write(f"{out_col_type}(")
        for out_row in range(a_r):
            out.write(f"{dot_name}(")
            write_mat_row("a", a_shape, out_row, out)
            out.write(", ")
            write_mat_col("b", b_shape, out_col, out)
            out.write(")")
            if out_row < a_r - 1:
                out.write(", ")
        if out_col_type == "f32":
            out.write(";\n")
        else:
            out.write(");\n")
    # Combine the columns into the output matrix
    if out_r == 1 and out_c == 1:
        out.write("    return col0;\n")
    else:
        out.write(f"    return {out_type}(")
        for out_col in range(b_c):
            out.write(f"col{out_col}")
            if out_col < b_c - 1:
                out.write(", ")
        out.write(");\n")
    out.write("}\n")

def write_vec(name, vec, out):
    r = vec.shape[0]
    vec_type = get_mat_type((r, 1))
    if name is not None:
        out.write(f"const {name}: {vec_type} = ")
    if r > 1:
        out.write(f"{vec_type}(")
    if r <= 4:
        for row in range(r):
            out.write(f"{vec[row]:.9f}")
            if row < r - 1:
                out.write(", ")
    else:
        num_vec4s = r // 4
        rem_vec_len = r % 4
        for i in range(num_vec4s):
            out.write("\n    ")
            write_vec(None, vec[i*4:(i+1)*4], out)
            if i < num_vec4s - 1 or rem_vec_len > 0:
                out.write(", ")
        if rem_vec_len > 0:
            out.write("\n    ")
            write_vec(None, vec[num_vec4s*4:], out)
    if r > 1:
        out.write(")")
    if name is not None:
        out.write(";\n")

def write_mat(name, mat, out):
    r, c = mat.shape
    if c == 1:
        write_vec(name, mat[:, 0], out)
    else:
        matType = get_mat_type((r, c))
        write_mat_defs(r, c, out)
        out.write(f"const {name}: {matType} = {matType}(\n")
        for col in range(c):
            write_vec(None, mat[:, col], out)
            if col < c - 1:
                out.write(",\n")
        out.write("\n);\n")



from alang.typs import array, struct
from alang.nodes import CodeOptions

def test_simple_struct_layout():
    # https://www.w3.org/TR/WGSL/#structure-member-layout
    s_a = struct(
        "A",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    l = s_a.layout
    fs = l.fields
    assert len(fs) == 4
    assert fs[0].triple == (0, 4, 4)
    assert fs[1].triple == (4, 4, 4)
    assert fs[2].triple == (8, 8, 8)
    assert fs[3].triple == (16, 4, 4)
    assert l.align == 8
    assert l.byte_size == 24

def test_nested_struct_layout():
    # https://www.w3.org/TR/WGSL/#structure-member-layout
    s_a = struct(
        "A",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    a = array(s_a, 3)
    s_b = struct(
        "B",
        ("a", "vec2f"),
        ("b", "vec3f"),
        ("c", "float"),
        ("d", "float"),
        ("e", s_a),
        ("f", "vec3f"),
        ("g", a),
        ("h", "int"),
    )
    l = s_b.layout
    fs = l.fields
    assert len(fs) == 8
    assert fs[0].triple == (0, 8, 8)
    assert fs[1].triple == (16, 16, 12)
    assert fs[2].triple == (28, 4, 4)
    assert fs[3].triple == (32, 4, 4)
    assert fs[4].triple == (40, 8, 24)
    assert fs[5].triple == (64, 16, 12)
    assert a.layout.element_stride == 24
    assert fs[6].triple == (80, 8, 72)
    assert fs[7].triple == (152, 4, 4)
    assert l.align == 16
    assert l.byte_size == 160

def test_struct_a():
    s = struct("Point", ("x", "int"), ("y", "int"))
    assert s.code.strip() == """
struct Point:
    x: int
    y: int
    """.strip()

def test_struct_wgsl():
    s = struct(
        "Ray",
        ("id", "int"),
        ("position", "vec3f"),
        ("direction", "vec3f"),
    )
    assert s.wgsl_code.strip() == """
struct Ray {
    id: i32,
    position: vec3<f32>,
    direction: vec3<f32>
}""".strip()

def test_runtime_sized():
    # https://www.w3.org/TR/WGSL/#buffer-binding-determines-runtime-sized-array-element-count
    point_light = struct(
        "PointLight",
        ("position", "vec3f"),
        ("color", "vec3f"),
    )
    light_storage = struct(
        "LightStorage",
        ("pointCount", "uint"),
        ("point", array(point_light)),
    )
    pl = point_light.layout
    assert pl.align == 16
    assert pl.byte_size == 32
    
    def test_buffer_size(buffer_size, expected_point_count):
        sl = light_storage.get_layout(buffer_size)
        fl = sl.fields[-1]
        assert sl.align == 16
        assert fl.num_elements == expected_point_count

    test_buffer_size(1024, 31)
    test_buffer_size(1025, 31)
    test_buffer_size(1039, 31)
    test_buffer_size(1040, 32)

def test_annotations():
    # https://www.w3.org/TR/WGSL/#structure-member-layout
    s_a = struct(
        "A",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    assert s_a.get_code("wgsl", CodeOptions(struct_annotations=True)).strip() == """
struct A {                                     //             align(8)  size(24)
    u: f32,                                    // offset(0)   align(4)  size(4)
    v: f32,                                    // offset(4)   align(4)  size(4)
    w: vec2<f32>,                              // offset(8)   align(8)  size(8)
    x: f32                                     // offset(16)  align(4)  size(4)
}
""".strip()



















import sys
import traceback




class Lexer :

  def __init__ (self) :
    pass



  def yylex (self) :
    return (0,None)


  def yyerror (self, msg) :
    s = msg
    
    sys.stderr.write(s+'\n')




YYACCEPT = 0

YYABORT = 1
YYERROR = 2




YYERRLAB = 3
YYNEWSTATE = 4
YYDEFAULT = 5
YYREDUCE = 6
YYERRLAB1 = 7
YYRETURN = 8



YYACTION = 10

LABELNAMES = (
"YYACCEPT",
"YYABORT",
"YYERROR",
"YYERRLAB",
"YYNEWSTATE",
"YYDEFAULT",
"YYREDUCE",
"YYERRLAB1",
"YYRETURN",
"YYGETTOKEN",
"YYACTION"
)


EOF = 0


YYEOF = 0
YYerror = 256
YYUNDEF = 257
INT_LITERAL = 258
NAME = 259
NEGATE = 260
INDENT = 261
DEDENT = 262
EOL = 263








def yy_pact_value_is_default_ (yyvalue) :
  return yyvalue == yypact_ninf_


def yy_table_value_is_error_ (yyvalue) :
  return yyvalue == yytable_ninf_


yypact_ = (
      36,   -45,    49,    41,    41,    41,     4,   -45,   -45,    63,
      41,     1,    31,    41,    -2,    17,    63,    18,    15,   -45,
      19,   -45,    36,    41,    41,    41,    41,    41,    63,    41,
      38,    28,    30,    41,   -45,    41,   -45,   -45,     6,     6,
      17,    17,    17,    63,    36,    46,   -45,    37,   -45,    43,
      36,   -45,    36,    48,    13,    36,   -45,    45,   -45
  )

yydefact_ = (
       0,    14,    15,     0,     0,     0,     0,     2,     4,     5,
       0,     0,     0,     0,    15,    22,    30,     0,    26,    28,
       0,     1,     0,     0,     0,     0,     0,     0,     6,     0,
       0,     0,     0,     0,    24,    27,    25,     3,    19,    18,
      20,    21,    23,     7,     8,    16,    17,     0,    29,     9,
      11,    16,     0,    12,     0,     0,    10,     0,    13
  )

yypgoto_ = (
     -45,   -44,   -19,   -45,    -3,     0,   -45,    29
  )

yydefgoto_ = (
       0,     6,     7,     8,     9,    17,    18,    19
  )

yytable_ = (
      15,    16,    16,    37,    21,    20,    29,    28,    54,    16,
      16,    57,    31,    32,    25,    33,    26,    13,    27,    22,
      38,    39,    40,    41,    42,    49,    43,    56,    22,    27,
      16,    53,    16,    47,     1,    14,    34,     3,    35,     1,
       2,    36,     3,    44,     1,    14,    45,     3,     4,    30,
      46,    50,     5,     4,    10,    51,    52,     5,     4,    58,
      22,    55,     5,     0,    48,    11,    12,     0,    13,    23,
      24,    25,     0,    26,     0,    27
  )

yycheck_ = (
       3,     4,     5,    22,     0,     5,     5,    10,    52,    12,
      13,    55,    12,    13,     8,    17,    10,    19,    12,    15,
      23,    24,    25,    26,    27,    44,    29,    14,    15,    12,
      33,    50,    35,    33,     3,     4,    18,     6,    23,     3,
       4,    22,     6,     5,     3,     4,    18,     6,    17,    18,
      20,     5,    21,    17,     5,    18,    13,    21,    17,    14,
      15,    13,    21,    -1,    35,    16,    17,    -1,    19,     6,
       7,     8,    -1,    10,    -1,    12
  )

yystos_ = (
       0,     3,     4,     6,    17,    21,    25,    26,    27,    28,
       5,    16,    17,    19,     4,    28,    28,    29,    30,    31,
      29,     0,    15,     6,     7,     8,    10,    12,    28,     5,
      18,    29,    29,    17,    18,    23,    22,    26,    28,    28,
      28,    28,    28,    28,     5,    18,    20,    29,    31,    26,
       5,    18,    13,    26,    25,    13,    14,    25,    14
  )

yyr1_ = (
       0,    24,    25,    25,    26,    26,    27,    27,    27,    27,
      27,    27,    27,    27,    28,    28,    28,    28,    28,    28,
      28,    28,    28,    28,    28,    28,    29,    29,    30,    30,
      31
  )

yyr2_ = (
       0,     2,     1,     3,     1,     1,     3,     4,     4,     5,
       8,     5,     6,     9,     1,     1,     4,     4,     3,     3,
       3,     3,     2,     3,     3,     3,     1,     2,     1,     3,
       1
  )


yytoken_number_ = (
       0,   256,   257,   258,   259,    61,    45,    43,    42,    64,
      47,   260,    94,   261,   262,   263,    95,    40,    41,    91,
      93,   123,   125,    44
  )

yytname_ = (
  "\"end of file\"", "error", "\"invalid token\"", "INT_LITERAL", "NAME",
  "'='", "'-'", "'+'", "'*'", "'@'", "'/'", "NEGATE", "'^'", "INDENT",
  "DEDENT", "EOL", "'_'", "'('", "')'", "'['", "']'", "'{'", "'}'", "','",
  "$accept", "stmts", "stmt", "define", "expr", "named_exprs",
  "named_expr_i", "named_expr", None
  )

yyrline_ = (
       0,    29,    29,    30,    34,    35,    39,    40,    41,    42,
      43,    44,    45,    46,    50,    51,    52,    53,    54,    55,
      56,    57,    58,    59,    60,    61,    65,    66,    70,    71,
      75
  )

yytranslate_table_ = (
       0,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
      17,    18,     8,     7,    23,     6,     2,    10,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     5,     2,     2,     9,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,    19,     2,    20,    12,    16,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,    21,     2,    22,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     1,     2,     3,     4,
      11,    13,    14,    15
  )

def yytranslate_ (t) :
  if (t >= 0 and t <= yyuser_token_number_max_) :
    return yytranslate_table_[t]
  else :
    return yyundef_token_


yylast_ = 75
yynnts_ = 8
yyempty_ = -2
yyfinal_ = 21
yyterror_ = 1
yyerrcode_ = 256
yyntokens_ = 24

yyuser_token_number_max_ = len(yytranslate_table_)-1 #e4_user_token_number_max
yyundef_token_ = 2 #e4_undef_token_number

yypact_ninf_ = -45
yytable_ninf_ = -1




class YYStack :

  def __init__ (self) :
    self.size = 16
    self.height = -1
    self.stateStack=[]
    self.valueStack=[]
    

  def push (self, state, value, ) :
    self.height += 1
    self.stateStack.append(state)
    self.valueStack.append(value)
    

  def pop (self, num) :
    if (num > 0) :
      for i in range(num) :
        self.valueStack.pop()
        self.stateStack.pop()
        
    self.height -= num

  def stateAt (self, i) :
    return self.stateStack[self.height - i]

  
  def valueAt (self, i) :
    return self.valueStack[self.height - i]

  def yyprint (self, out) :
    out.write ("Stack now")
    for x in self.stateStack[:] :
        out.write (' ')
        out.write (str(x))
    out.write ('\n')




class YYVAL:
  def __init__(self,yyval): self.yyval = yyval


class YYParser :
  bisonVersion = "30802"

  bisonSkeleton = "src/alang/langs/py.skel"





  def getErrorVerbose(self) :
    return self.yyErrorVerbose

  def setErrorVerbose(self, verbose) :
    self.yyErrorVerbose = verbose


  def getDebugStream (self) :
    return self.yyDebugStream

  def setDebugStream(self, s) :
    self.yyDebugStream = s

  def getDebugLevel (self) :
    return self.yydebug

  def setDebugLevel (self, level) :
    self.yydebug = level



  def __init__  (self, yylexer):

    self.yylexer = yylexer


    self.yyDebugStream = sys.stderr
    self.yydebug = 0
    self.yyerrstatus_ = 0


    self.yyErrorVerbose = True
    self.yyswitch_init()




  def yyswitch_init(self):
    self.yyswitch = {}








  def yyaction (self, yyn, yystack, yylen, yyerror) :
    yylval = None
    

    if (yylen > 0) :
      yyval = yystack.valueAt (yylen - 1)
    else :
      yyval = yystack.valueAt (0)

    self.yy_reduce_print (yyn, yystack)


    if (yyn in self.yyswitch) :
      action = self.yyswitch[yyn]
      yyvalp = YYVAL(yyval)
      status = action(yyvalp, yyn, yystack, yylen, yyerror)
      yyval = yyvalp.yyval
      if(status != None) : return status
    else: # no such action index; ignore
      pass

    self.yy_symbol_print ("-> $$ =",
                          yyr1_[yyn], yyval)

    yystack.pop (yylen)
    yylen = 0

    yyn = yyr1_[yyn]
    tmp = yyntokens_ # quote problem
    yystate = yypgoto_[yyn - tmp] + yystack.stateAt (0)
    if (0 <= yystate
        and yystate <= yylast_
        and yycheck_[yystate] == yystack.stateAt (0)) :
      yystate = yytable_[yystate]
    else :
      yystate = yydefgoto_[yyn - tmp]

    yystack.push (yystate, yyval)
    return YYNEWSTATE


  def yy_reduce_print (self, yyrule, yystack) :
    if (self.yydebug == 0) :
      return

    yylno = yyrline_[yyrule]
    yynrhs = yyr2_[yyrule]
    self.yycdebug ("Reducing stack by rule " + str(yyrule - 1)
               + " (line " + str(yylno) + "), ")

    for yyi in range(yynrhs) :
      self.yy_symbol_print ("   $" + str(yyi + 1) + " =",
                       yystos_[yystack.stateAt(yynrhs - (yyi + 1))],
                       (yystack.valueAt (yynrhs-(yyi + 1))))





  def  parse (self) :
    

    yychar = yyempty_
    yytoken = 0

    yyn = 0
    yylen = 0
    yystate = 0
    yystack = YYStack ()
    label = YYNEWSTATE

    yynerrs_ = 0
    
    yylval = None





    self.yycdebug ("Starting parse\n")
    self.yyerrstatus_ = 0
    yystack.push (yystate,
                           yylval )





    while True :

      if label == YYNEWSTATE : # case YYNEWSTATE
        self.yycdebug ("Entering state " + str(yystate) + '\n')
        if (self.yydebug > 0) :
          yystack.yyprint (self.yyDebugStream)

        if (yystate == yyfinal_) :
          return True

        tmp = yystate
        yyn = yypact_[tmp]
        if (yy_pact_value_is_default_ (yyn)) :
            label = YYDEFAULT
            continue; # break switch


        if (yychar == yyempty_) :
          self.yycdebug ("Reading a token: ")
          yychar, yylval = self.yylexer.yylex()
          

        if (yychar <= EOF) :
          yychar = EOF
          yytoken = EOF
          self.yycdebug ("Now at end of input.\n")
        else :
          yytoken = yytranslate_ (yychar)
          self.yy_symbol_print ("Next token is",
                                yytoken,
                                yylval
                                
                                )

        yyn += yytoken
        tmp = yyn # Quote problem
        if (yyn < 0 
            or yylast_ < yyn
            or yycheck_[tmp] != yytoken) :
          label = YYDEFAULT

        elif (yytable_[tmp] <= 0) :
          yyn = yytable_[tmp]
          if (yy_table_value_is_error_ (yyn)) :
            label = YYERRLAB
          else :
            yyn = -yyn
            label = YYREDUCE
        else :
          tmp = yyn # Quote problem
          yyn = yytable_[tmp]
          self.yy_symbol_print ("Shifting",
                                yytoken,
                                yylval)

          yychar = yyempty_

          if (self.yyerrstatus_ > 0) :
              self.yyerrstatus_ -= 1

          yystate = yyn
          yystack.push (yystate,             yylval)
          label = YYNEWSTATE

      elif label == YYDEFAULT : #case YYDEFAULT
        tmp = yystate # Quote problem
        yyn = yydefact_[tmp]
        if (yyn == 0) :
          label = YYERRLAB
        else :
          label = YYREDUCE

      elif label == YYREDUCE : #case YYREDUCE
        tmp = yyn # Quote problem
        yylen = yyr2_[tmp]
        label = self.yyaction (yyn,         yystack, yylen, self.yylexer.yyerror)
        yystate = yystack.stateAt (0)

      elif label == YYERRLAB: #case YYERRLAB
        if (self.yyerrstatus_ == 0) :
          yynerrs_ += 1
          if (yychar == yyempty_) :
            yytoken = yyempty_
          tmp = self.yysyntax_error (yystate, yytoken)
          self.yyerror (tmp)

        
        if (self.yyerrstatus_ == 3) :

          if (yychar <= EOF) :
            if (yychar == EOF) :
              return False
          else :
            yychar = yyempty_

        label = YYERRLAB1

      elif label == YYERROR : #case YYERROR
        
        yystack.pop (yylen)
        yylen = 0
        yystate = yystack.stateAt (0)
        label = YYERRLAB1

      elif label == YYERRLAB1 : #case YYERRLAB1
        self.yyerrstatus_ = 3 # Each real token shifted decrements this.
        while True :
          tmp = yystate # Quote problem
          yyn = yypact_[tmp]
          if ( not yy_pact_value_is_default_ (yyn)) :
            yyn += yyterror_
            tmp = yyn # Quote problem
            if (0 <= yyn and yyn <= yylast_ \
                and yycheck_[tmp] == yyterror_) :
              yyn = yytable_[tmp]
              if (0 < yyn) :
                break # leave while loop

          if (yystack.height == 0) :
            return False

          
          yystack.pop (1)
          yystate = yystack.stateAt (0)
          if (self.yydebug > 0) :
            yystack.yyprint (self.yyDebugStream)

        if (label == YYABORT) :
          continue # Leave the switch.


        tmp = yyn
        self.yy_symbol_print ("Shifting", yystos_[tmp],
                         yylval)

        yystate = yyn
        yystack.push (yyn, yylval               )
        label = YYNEWSTATE
        continue # leave the switch

      elif label == YYACCEPT : # case YYACCEPT
        return True

      elif label == YYABORT: # case YYABORT
        return False

      else :
        assert False, "Unknown State:" + str(label)










  def yyerror (self, msg) :
    self.yylexer.yyerror (msg)

  def yycdebug (self, s) :
    if (self.yydebug > 0) :
      self.yyDebugStream.write (s+'\n')


  def yytnamerr_ (self, yystr) :
    yyr = ""
    if (yystr[0] == '"')  :
      l = len(yystr)
      i = 1
      while (True) :
        if (i >= l) : break
        c = yystr[i]
        if(c == "'" or c == ',') :
          continue
        if( c == '"'):
          return yyr
        if(c == '\\') :
          i += 1
          c = yystr[i]
          if(c != '\\') :
            break
        yyr = yyr + c
        i += 1
    elif (yystr ==  "$end") :
      return "end of input"
    return yystr;



  def yy_symbol_print (self, s, yytype, yyvaluep ) :
    if (self.yydebug > 0) :
      tag = " nterm "
      if (yytype < yyntokens_) :
        tag = " token "
      if (yyvaluep is None) :
        vps = "None"
      else :
        vps = str(yyvaluep)
      tname = yytname_[yytype]
      line = s + tag + tname
      line += " ("
      
      line += vps
      line += ')'
      self.yycdebug (line)

  def yysyntax_error (self, yystate, tok) :
  
    if (self.yyErrorVerbose) :

      if (tok  != yyempty_) :
        res = "syntax error, unexpected "
        res += (self.yytnamerr_ (yytname_[tok]))
        tmp = yystate
        yyn = yypact_[tmp]
        if ( not yy_pact_value_is_default_ (yyn)) :
          yyxbegin = 0
          if (yyn < 0) :
            yyxbegin =  - yyn
          yychecklim = yylast_ - yyn + 1
          yyxend = yychecklim
          if (yychecklim >= yyntokens_) :
            yyxend = yyntokens_
          count = 0
          for x in range(yyxbegin,yyxend) :
            tmp = yyn
            if (yycheck_[x + tmp] == x and x != yyterror_
                and  not yy_table_value_is_error_ (yytable_[x + tmp])) :
              count += 1
          if (count < 5) :
            count = 0
            for x in range(yyxbegin,yyxend) :
              tmp = yyn
              if (yycheck_[x + tmp] == x and x != yyterror_
                  and  not yy_table_value_is_error_ (yytable_[x + tmp])) :
                if (count == 0) :
                  res += ", expecting "
                else :
                  res += " or "
                count += 1
                res += (self.yytnamerr_ (yytname_[x]))
        return str(res)

    return "syntax error"

  





from typing import Optional, TextIO, Union

from alang.langs.language import Language, register_language
from alang.langs.c import CWriter
import alang.typs as typs

class GLSLWriter(CWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)

    def get_type_name(self, t: typs.Type) -> str:
        if t is None:
            return "void"
        elif isinstance(t, typs.Integer):
            if t.bits == 32:
                if t.signed:
                    return "int"
                else:
                    return "uint"
            else:
                self.error(f"Integer type `{t.name}` not supported")
                return "int" if t.signed else "uint"
        elif isinstance(t, typs.Float):
            if t.bits == 32:
                return "float"
            elif t.bits == 64:
                return "double"
            else:
                self.error(f"Float type `{t.name}` not supported")
                return "float"
        elif isinstance(t, typs.Vector):
            n = t.size
            if n > 4 or n < 2:
                self.error(f"Invalid vector size: {n}")
                n = 4
            element_type_name = self.get_type_name(t.element_type).name
            if element_type_name == "float":
                return f"vec{n}"
            elif element_type_name == "double":
                return f"dvec{n}"
            elif element_type_name == "int":
                return f"ivec{n}"
            elif element_type_name == "uint":
                return f"uvec{n}"
        else:
            return t.name

class GLSLLanguage(Language):
    def __init__(self):
        super().__init__("glsl")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> GLSLWriter: # type: ignore
        return GLSLWriter(out, options)

glsl_lang = GLSLLanguage()
register_language(glsl_lang)



from typing import Optional

from alang.langs.writer import CodeWriter
from alang.nodes import Expression, Node, NodeAttr, NodeLink, NodeLinks, NodeType

import alang.compiler as compiler
import alang.funcs as funcs
import alang.stmts as stmts
import alang.typs as typs

class Name(Expression):
    name = NodeAttr()
    def __init__(self, name: str):
        super().__init__(NodeType.NAME)
        self.name = name
        self.resolved_node: Node = None
    def resolve_type(self, diags: compiler.Diagnostics) -> typs.Type:
        if self.resolved_node is None:
            return None
        return self.resolved_node.resolved_type

class BinopOp:
    def __init__(self, name: str, op: str, precedence: int = 0, left_assoc: bool = True):
        self.name = name
        self.op = op
        self.precedence = precedence
        self.left_assoc = left_assoc
    def __str__(self):
        return self.op
    def __repr__(self):
        return repr(self.op)

bops = [
    BinopOp("add", "+"),
    BinopOp("sub", "-"),
    BinopOp("mul", "*"),
    BinopOp("matmul", "@"),
    BinopOp("div", "/"),
    BinopOp("mod", "%"),
    BinopOp("eq", "=="),
    BinopOp("ne", "!="),
    BinopOp("lt", "<"),
    BinopOp("le", "<="),
    BinopOp("gt", ">"),
    BinopOp("ge", ">="),
    BinopOp("and", "&&"),
    BinopOp("or", "||"),
    BinopOp("xor", "^"),
    BinopOp("shl", "<<"),
    BinopOp("shr", ">>"),
    BinopOp("band", "&"),
    BinopOp("bor", "|"),    
]
bop_from_name = {bop.name: bop for bop in bops}
bop_from_op = {bop.op: bop for bop in bops}
    
class Binop(Expression):
    left = NodeLink()
    right = NodeLink()
    operator = NodeAttr()
    def __init__(self, left: Expression, operator: str, right: Expression):
        super().__init__(NodeType.BINOP)
        if operator in bop_from_name:
            self.operator = bop_from_name[operator]
        elif operator in bop_from_op:
            self.operator = bop_from_op[operator]
        else:
            raise Exception(f"Unknown operator: {operator}")
        self.left = parse_expr(left)
        self.right = parse_expr(right)
    def resolve_type(self, diags: compiler.Diagnostics) -> typs.Type:
        lt: typs.Type = self.left.resolved_type
        rt: typs.Type = self.right.resolved_type
        if lt is None or rt is None:
            return None
        if not lt.is_algebraic:
            diags.error(f"Not an algebraic type", self.left)
            return None
        if not rt.is_algebraic:
            diags.error(f"Not an algebraic type", self.right)
            return None
        opn = self.operator.name
        if opn == "matmul":
            if lt.is_tensor and rt.is_tensor:
                return lt @ rt
            else:
                return None
        elif opn in ["add", "sub", "mul", "div", "mod"]:
            if lt.is_floatish:
                return lt
            elif rt.is_floatish:
                return rt
            else:
                return lt
        elif opn in ["shl", "shr", "band", "bor"]:
            return lt
        else:
            diags.error(f"Binary operator {self.operator.op} type resolution not supported")
            return lt
    def add(self, other: Expression) -> Expression:
        other = parse_expr(other)
        if other.node_type == NodeType.CONSTANT:
            if other.is_zero:
                return self
            if self.operator.name == "add" and self.right.node_type == NodeType.CONSTANT:
                return Binop(self.left, "add", self.right + other)
        return Binop(self, "add", other)
    def mul(self, other: Expression) -> Expression:
        other = parse_expr(other)
        if other.node_type == NodeType.CONSTANT:
            if other.is_zero:
                return other
            if other.is_one:
                return self
            if self.operator.name == "mul" and self.right.node_type == NodeType.CONSTANT:
                return Binop(self.left, "mul", self.right * other)
        return Binop(self, "mul", other)
    def get_support_lib_function_name(self) -> str:
        if self.operator.name == "matmul":
            lt: typs.Tensor = self.left.resolved_type
            rt: typs.Tensor = self.right.resolved_type
            if lt is not None and rt is not None and lt.is_tensor and rt.is_tensor:
                return f"mul_{lt.name}_{rt.name}"
        return None
    def get_support_definitions(self, defs: compiler.SupportDefinitions):
        if self.operator.name == "matmul":
            lt: typs.Tensor = self.left.resolved_type
            rt: typs.Tensor = self.right.resolved_type
            if lt is not None and rt is not None and lt.is_tensor and rt.is_tensor:
                name = self.get_support_lib_function_name()
                if name is not None and defs.needs(name):
                    from alang.stmts import Set
                    from alang.exprs import parse_expr
                    ot = lt @ rt
                    f = funcs.Function(name, ot, ("a", lt), ("b", rt))
                    f.var("o", ot)
                    inner_count = rt.shape[0]
                    inner_add = None
                    o_index_expr = Index("o", ot.get_flat_index(["out_r", "out_c"]))
                    for inner_i in range(inner_count):
                        a_index = Index("a", lt.get_flat_index(("out_r", inner_i)))
                        b_index = Index("b", rt.get_flat_index((inner_i, "out_c")))
                        mul = a_index * b_index
                        if inner_add is None:
                            inner_add = mul
                        else:
                            inner_add = inner_add + mul
                    c_loop = stmts.Loop("out_c", ot.shape[1], Set(o_index_expr, inner_add))
                    f.loop("out_r", ot.shape[0], c_loop).ret("o")
                    defs.add(name, [f])
        return None
    
class Constant(Expression):
    value = NodeAttr()
    def __init__(self, value: object):
        super().__init__(NodeType.CONSTANT)
        self.value = value
    @property
    def is_zero(self) -> bool:
        return (self.value == 0)
    @property
    def is_one(self) -> bool:
        return (self.value == 1)
    def resolve_type(self, diags: compiler.Diagnostics) -> typs.Type:
        v = self.value
        if v is None:
            return None
        if isinstance(v, int):
            return typs.int_type
        if isinstance(v, float):
            return typs.float_type
        return None
    def mul(self, other: Expression) -> Expression:
        if self.is_zero:
            return self
        other = parse_expr(other)
        if other.node_type == NodeType.CONSTANT:
            if other.is_zero:
                return other
            return Constant(self.value * other.value)
        return Binop(self, "mul", other)
    def add(self, other: Expression) -> Expression:
        other = parse_expr(other)
        if self.is_zero:
            return other
        if other.node_type == NodeType.CONSTANT:
            if other.is_zero:
                return self
            return Constant(self.value + other.value)
        return Binop(self, "add", other)
    def sub(self, other: Expression) -> Expression:
        other = parse_expr(other)
        if other.node_type == NodeType.CONSTANT:
            if other.is_zero:
                return self
            return Constant(self.value - other.value)
        return Binop(self, "sub", other)
    def div(self, other: Expression) -> Expression:
        if self.is_zero:
            return self
        other = parse_expr(other)
        if other.node_type == NodeType.CONSTANT:
            return Constant(self.value / other.value)
        return Binop(self, "div", other)

class Funcall(Expression):
    func = NodeLink()
    args = NodeLinks()
    def __init__(self, func: Expression, args: list[Expression] = []):
        super().__init__(NodeType.FUNCALL)
        self.func = parse_expr(func)
        for a in args:
            self.link(parse_expr(a), "args")
    def resolve_type(self, diags: compiler.Diagnostics) -> typs.Type:
        f = self.func
        if f is None:
            return None
        ft = f.resolved_type
        if ft is None:
            return None
        return ft.return_type
    def write(self, writer: CodeWriter):
        f = self.func
        if f is None:
            return
        f.write(writer)
        writer.write("(")
        for i, arg in enumerate(self.args):
            if i > 0:
                writer.write(", ")
            arg.write(writer)
        writer.write(")")

class Index(Expression):
    base = NodeLink()
    ranges = NodeLinks()
    def __init__(self, base: Expression, *ranges: list[Expression]):
        super().__init__(NodeType.INDEX)
        self.base = parse_expr(base)
        for r in ranges:
            self.link(parse_expr(r), "ranges")
    def resolve_type(self, diags: compiler.Diagnostics) -> typs.Type:
        bt: typs.Type = self.base.resolved_type
        if bt is None:
            return None
        if not bt.is_indexable:
            diags.error(f"Indexing not supported for type {bt}")
            return None
        return bt.element_type

def parse_expr(expr: object, context: Optional[Node] = None) -> Node:
    if expr is None:
        return None
    elif isinstance(expr, Node):
        return expr
    elif isinstance(expr, int):
        return Constant(expr)
    elif isinstance(expr, float):
        return Constant(expr)
    else:
        from alang.langs.a import a_lang
        return a_lang.parse_expr(expr)



# from alang.langs.language import Language, get_language as get_language_x

def get_language(lang_name: str) -> "Language": # type: ignore
    import alang.langs.html_dev
    import alang.langs.a
    import alang.langs.c
    import alang.langs.glsl
    import alang.langs.js
    import alang.langs.metal
    import alang.langs.wgsl
    import alang.langs.language
    return alang.langs.language.get_language(lang_name)



from typing import Optional, TextIO, Union

from alang.langs.language import Language, register_language
from alang.langs.writer import CodeWriter
import alang.exprs as exprs
import alang.typs as typs
import alang.funcs as funcs
import alang.stmts as stmts

class JSWriter(CodeWriter):
    def __init__(self, out: Union[str, TextIO], options: Optional["CodeOptions"]): # type: ignore
        super().__init__(out, options)

    def write_alias(self, b: "Alias"): # type: ignore
        pass # No types in JS

    def write_binop(self, b: exprs.Binop):
        support_name = b.get_support_lib_function_name()
        if support_name is not None:
            self.write(support_name)
            self.write("(")
            self.write_expr(b.left)
            self.write(", ")
            self.write_expr(b.right)
            self.write(")")
        else:
            self.write("(")
            self.write_expr(b.left)
            self.write(" ")
            self.write(b.operator.op)
            self.write(" ")
            self.write_expr(b.right)
            self.write(")")

    def write_block(self, b: stmts.Block):
        self.write("{\n")
        self.indent()
        for s in b.statements:
            self.write_node(s)
        self.dedent()
        self.write("}\n")

    def write_constant(self, c: "Constant"): # type: ignore
        self.write(repr(c.value))

    def write_expr_stmt(self, e: stmts.ExprStmt):
        self.write_expr(e.expression)
        self.write(";\n")

    def write_function(self, f: funcs.Function):
        self.write_js_function(f)
        s = self.get_func_stage(f)
        if s is not None:
            self.write_gpu_function(f, s)

    def write_js_function(self, f: funcs.Function):
        self.write(f"function {f.name}(")
        for i, param in enumerate(f.parameters):
            if i > 0:
                self.write(", ")
            self.write(f"{param.name}/*: {self.get_typed_name(param.parameter_type)}*/")
        self.write(") ")
        self.write_block(f)

    def write_gpu_function(self, original_f: funcs.Function, stage_and_f: tuple[str, funcs.Function]):
        stage, f = stage_and_f
        pts = [(p.resolved_type or p.parameter_type) for p in f.parameters]
        self.write(f"class {original_f.name}GPU {{\n")
        self.indent()
        self.writeln(f"constructor(device, shaderModule) {{")
        self.indent()
        self.writeln("this.device = device;")
        self.writeln(f"console.log('Creating GPU function {f.name}');")
        for i, param in enumerate(f.parameters):
            self.writeln(f"self.{param.name} = null; // {self.get_typed_name(pts[i])}")
        self.writeln("this.computePipeline = device.createComputePipeline({")
        self.indent()
        self.writeln(f"label: '{f.name}_pipeline',")
        self.writeln("layout: 'auto',")
        self.writeln("compute: {")
        self.indent()
        self.writeln("module: shaderModule,")
        self.writeln(f"entryPoint: '{f.name}'")
        self.dedent()
        self.writeln("}")
        self.dedent()
        self.writeln("});")
        self.writeln("const bindGroupLayout = this.computePipeline.getBindGroupLayout(0);")
        self.writeln("console.log('Created GPU function pipeline', this.computePipeline);")
        self.writeln("this.bindGroup = device.createBindGroup({")
        self.indent()
        self.writeln("layout: bindGroupLayout,")
        self.writeln("entries: [")
        self.indent()
        for i, param in enumerate(f.parameters):
            break
            self.writeln("{")
            self.indent()
            self.writeln(f"binding: {i},")
            self.writeln(f"resource: this.{param.name}.gpuBuffer")
            self.dedent()
            self.writeln("},")
        self.dedent()
        self.writeln("]")
        self.dedent()
        self.writeln("});")
        self.writeln("console.log('Created GPU function bind group', this.bindGroup);")
        self.dedent()
        self.writeln("}")
        self.write("encode() {")
        self.indent()
        self.writeln(f"console.log('Encoding GPU function {f.name}');")
        self.writeln("const commandEncoder = this.device.createCommandEncoder();")
        self.writeln(f"const passEncoder = commandEncoder.beginComputePass({{label:'{f.name}_pass'}});")
        self.writeln("passEncoder.setPipeline(this.computePipeline);")
        self.writeln("passEncoder.setBindGroup(0, this.bindGroup);")
        self.writeln("passEncoder.dispatchWorkgroups(1);")
        self.writeln("passEncoder.end();")
        self.writeln(f"return commandEncoder.finish({{label:'{f.name}_command'}});")
        self.dedent()
        self.writeln("}")
        self.write("exec() {")
        self.indent()
        self.writeln(f"console.log('Executing GPU function {f.name}');")
        self.writeln("const commandBuffer = this.encode();")
        self.writeln("console.log('Executing GPU function command buffer', commandBuffer);")
        self.writeln("this.device.queue.submit([commandBuffer]);")
        self.dedent()
        self.writeln("}")
        self.dedent()
        self.write("}\n")

    def write_index(self, i: "Index"): # type: ignore
        self.write_expr(i.base)
        self.write("[")
        for ri, r in enumerate(i.ranges):
            if ri > 0:
                self.write(", ")
            if r is None:
                self.write(":")
            else:
                self.write_expr(r)
        self.write("]")

    def write_loop(self, f: stmts.Loop):
        self.write("for (let")
        self.write(f" {f.var} = 0; {f.var} < ")
        self.write_expr(f.count)
        self.write(f"; {f.var}++) {{\n")
        self.indent()
        for s in f.statements:
            self.write_node(s)
        self.dedent()
        self.write("}\n")

    def write_line_comment(self, comment: str):
        self.write("// ")
        self.write(comment)
        self.write("\n")

    def write_set(self, s: stmts.Set):
        self.write_expr(s.target)
        self.write(" = ")
        self.write_expr(s.value)
        self.write(";\n")

    def write_struct(self, s: typs.Struct):
        fs: list[typs.Field] = s.fields
        sl = s.layout
        n = len(fs)
        anno_col = 46
        anno = self.options.struct_annotations
        def write_anno(o, a, s, text_len):
            if anno:
                tab = " " * max(anno_col - text_len, 1)
                o_text = ""
                if o is not None:
                    o_text = f"offset({o})"
                a_text = f"align({a})"
                self.write(f"{tab}// {o_text.ljust(12)}{a_text.ljust(10)}size({s})")
        self.write(f"class {s.name} {{")
        write_anno(None, sl.align, sl.byte_size, len(s.name) + 8)
        is_scalar_array = [False] * n
        scalar_array_num_elements = [0] * n
        self.write("\n")
        self.write(f"    constructor(buffer, byteOffset, byteLength) {{\n")
        self.write(f"        byteOffset = byteOffset || 0;\n")
        self.write(f"        byteLength = byteLength || {sl.byte_size};\n")
        self.write(f"        if (byteLength < {sl.byte_size}) throw new Error(`Buffer too small. \"{s.name}\" requires at least {sl.byte_size} bytes, got ${{byteLength}}`);\n")
        self.write(f"        if (buffer instanceof ArrayBuffer) {{\n")
        self.write(f"            this.buffer = buffer;\n")
        self.write(f"        }} else {{\n")
        self.write(f"            this.buffer = new ArrayBuffer(byteLength);\n")
        self.write(f"            byteOffset = 0;\n")
        self.write(f"        }}\n")
        self.write(f"        if (byteOffset + byteLength > this.buffer.byteLength) throw new Error(`Buffer overflow. \"{s.name}\" requires ${{byteLength}} bytes starting at ${{byteOffset}}, but the buffer is only ${{this.buffer.byteLength}} bytes long`);\n")
        self.write(f"        this.view = new DataView(this.buffer, byteOffset, byteLength);\n")
        self.write(f"        this.byteLength = byteLength;\n")
        self.write(f"        this.gpuBuffer = null;\n")
        self.write(f"        this.isDirty = false;\n")
        self.write(f"        this.dirtyBegin = 0;\n")
        self.write(f"        this.dirtyEnd = 0;\n")
        for i, field in enumerate(fs):
            field_type: typs.Type = field.field_type
            if field_type.is_array:
                is_scalar_array[i] = field_type.element_type.is_scalar
                scalar_array_num_elements[i] = field_type.num_elements
            elif field_type.is_tensor:
                is_scalar_array[i] = field_type.element_type.is_scalar
                scalar_array_num_elements[i] = field_type.num_elements
            elif field_type.is_vector:
                is_scalar_array[i] = True
                scalar_array_num_elements[i] = field_type.size
            if is_scalar_array[i]:
                element_type_name = self.get_typed_name(field_type.element_type)
                self.write(f"        this.{field.name}Array = new {element_type_name}Array(this.buffer, byteOffset + {sl.fields[i].offset}, {scalar_array_num_elements[i]});\n")
        self.write(f"    }}\n")
        self.write(f"    dirty(begin, end) {{\n")
        self.write(f"        if (begin === undefined || end === undefined) {{ begin = 0; end = this.byteLength; }}\n")
        self.write(f"        if (this.isDirty) {{ this.dirtyBegin = Math.min(this.dirtyBegin, begin); this.dirtyEnd = Math.max(this.dirtyEnd, end); }}\n")
        self.write(f"        else {{ this.dirtyBegin = begin; this.dirtyEnd = end; }}\n")
        self.write(f"        this.isDirty = true;\n")
        self.write(f"    }}\n")
        self.write(f"    createGPUBuffer(device, usage) {{\n")
        self.write(f"        usage = usage || (GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_DST);\n")
        self.write(f"        this.gpuBuffer = device.createBuffer({{ size: Math.max(this.byteLength, 256), usage: usage, label: \"{s.name}\", mappedAtCreation: false }});\n")
        self.write(f"        device.queue.writeBuffer(this.gpuBuffer, 0, this.buffer, this.view.byteOffset, this.byteLength);\n")
        self.write(f"        this.isDirty = false;\n")
        self.write(f"        return this.gpuBuffer;\n")
        self.write(f"    }}\n")
        for i, field in enumerate(fs):
            field_type = field.field_type
            if field_type.is_scalar:
                field_type_name = self.get_typed_name(field_type)
                self.write(f"    get {field.name}() {{ return this.view.get{field_type_name}({sl.fields[i].offset}); }}\n")
                self.write(f"    set {field.name}(value) {{ return this.view.set{field_type_name}({sl.fields[i].offset}, value); this.dirty({sl.fields[i].offset}, {sl.fields[i].offset + sl.fields[i].byte_size}); }}\n")
            elif is_scalar_array[i]:
                self.write(f"    get {field.name}() {{ return this.{field.name}Array; }}\n")
                self.write(f"    set {field.name}(value) {{ this.{field.name}Array.set(value); this.dirty({sl.fields[i].offset}, {sl.fields[i].offset + sl.fields[i].byte_size}); }}\n")
            else:
                raise NotImplementedError(f"Field type {field_type} not implemented")
        self.write("}\n")

    def write_return(self, s: stmts.Return):
        self.write(f"return")
        rv = s.value
        if rv is not None:
            self.write(" ")
            self.write_expr(rv)
        self.write(";\n")

    def write_name(self, s: exprs.Name):
        self.write(s.name)

    def get_typed_name(self, t: typs.Type) -> str:
        if isinstance(t, typs.Integer):
            tn = t.name
            if tn == "sbyte":
                return "Int8"
            elif tn == "byte":
                return "UInt8"
            elif tn == "int":
                return "Int32"
            elif tn == "uint":
                return "UInt32"
            elif tn == "long":
                return "BigInt64"
            elif tn == "ulong":
                return "BigUint64"
            else:
                raise ValueError(f"Invalid integer type: {tn}")
        elif isinstance(t, typs.Float):
            tn = t.name
            if tn == "half":
                return "Float16"
            elif tn == "float":
                return "Float32"
            elif tn == "double":
                return "Float64"
            else:
                raise ValueError(f"Invalid float type: {tn}")
        elif isinstance(t, typs.Vector):
            return self.get_typed_array_name(t.element_type)
        else:
            return t.name
        
    def get_typed_array_name(self, t: typs.Type) -> str:
        return f"{self.get_typed_name(t)}Array"

    def get_js_name(self, t: typs.Type) -> str:
        if isinstance(t, typs.Integer):
            return "number"
        elif isinstance(t, typs.Float):
            return "number"
        elif isinstance(t, typs.Vector):
            return self.get_typed_array_name(t.element_type)
        elif isinstance(t, typs.Array):
            raise NotImplementedError("Arrays not implemented")
        else:
            return t.name

class JSLanguage(Language):
    def __init__(self):
        super().__init__("js")

    def open_writer(self, out: Union[str, TextIO], options: Optional["CodeOptions"]) -> JSWriter: # type: ignore
        return JSWriter(out, options)

js_lang = JSLanguage()
register_language(js_lang)



import os

from alang.nodes import define, CodeOptions, Module
from alang.langs import get_language
from alang.typs import struct

html_lang = get_language("html")

tests_dir = os.path.dirname(os.path.abspath(__file__))
proj_dir = os.path.dirname(tests_dir)
test_out_dir = os.path.join(proj_dir, "test_out")
os.makedirs(test_out_dir, exist_ok=True)

def write_standalone_html(mod_name, node):
    path = os.path.join(test_out_dir, f"{mod_name}.html")
    mod = node
    if mod.node_type != "module":
        mod = Module(mod_name)
        mod.append_any(node)
    mod.write_code(path, "html", CodeOptions(standalone=True, auto_entry_points=True))
    with open(path, "r") as f:
        return f.read()

def test_just_struct():
    s_a = struct(
        "TestStruct",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    assert s_a.html_code.strip().startswith("<h2>TestStruct</h2>")

def test_standalone_struct():
    s_a = struct(
        "TestStruct",
        ("u", "float"),
        ("v", "float"),
        ("w", "vec2f"),
        ("x", "float"),
    )
    html = write_standalone_html("test_standalone_struct", s_a)
    assert html.index("<h2>TestStruct</h2>") > 0

def test_standalone_function():
    f = define("f", ("x", "int")).ret("x")
    html = write_standalone_html("test_standalone_function", f)
    assert html.index("<h2>f</h2>") > 0
